from processing.common.AggregatedStorage import AggregatedStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_no_aggregated import print_no_aggregated


def validate_aggregated(action):
    """The aggregated data validator."""

    def decorator(
        storage: Storage,
        callback: MenuCallback,
    ):
        if not AggregatedStorage.exists(storage):
            print_no_aggregated()
            invoke_menu(storage, callback)
            return

        action(storage, callback)

    return decorator



import sys


def deep_getsizeof(ob):
    size = sys.getsizeof(ob)

    if isinstance(ob, (list, tuple, set)):
        for element in ob:
            size += deep_getsizeof(element)

    if isinstance(ob, dict):
        for k, v in ob.items():
            size += deep_getsizeof(k)
            size += deep_getsizeof(v)

    return size



from typing import List, Tuple

from processing.utils.validate_int import validate_int


class BandConfig:
    index: int
    name: str
    low: int
    high: int

    def __init__(
        self,
        index: int,
        name: str,
        low: int,
        high: int,
    ) -> None:
        validate_int(low)
        validate_int(high)

        self.index = index
        self.name = name
        self.low = low
        self.high = high

    @staticmethod
    def flatten(
        bands: List["BandConfig"],
    ) -> Tuple[List[str], List[int], List[int]]:
        names = [b.name for b in bands]
        lows = [b.low for b in bands]
        highs = [b.high for b in bands]

        return names, lows, highs

    @staticmethod
    def reconstruct(
        names: List[str],
        lows: List[int],
        highs: List[int],
    ) -> List["BandConfig"]:
        bands = []

        for index, name in enumerate(names):
            band = BandConfig(
                index=index,
                name=name,
                low=lows[index],
                high=highs[index],
            )

            bands.append(band)

        return bands



from typing import Dict, List

from processing.config.trajectories.TrajectoryConfig import TrajectoryConfig

PackedTrajectories = Dict[str, Dict[str, List[TrajectoryConfig]]]


def pack_trajectories(
    trajectories: List[TrajectoryConfig],
) -> PackedTrajectories:
    packed_trajectories: PackedTrajectories = {}

    for trajectory in trajectories:
        label_property = trajectory.label_property
        label_value = trajectory.label_value

        if label_property not in packed_trajectories:
            packed_trajectories[label_property] = {}

        if label_value not in packed_trajectories[label_property]:
            packed_trajectories[label_property][label_value] = []

        packed_trajectories[label_property][label_value].append(trajectory)

    return packed_trajectories



import numpy as np

from processing.common.AggregatedStorage import AggregatedStorage
from processing.common.TimelineWalker import TimelineWalker
from processing.config.bands.BandStorage import BandStorage
from processing.config.extractors.ExtractorStorage import ExtractorStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.config.settings.SettingsStorage import SettingsStorage
from processing.config.sites.SiteStorage import SiteStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.create_timelines import create_timelines
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.print_extractors import print_extractors
from processing.utils.validate_configuration import validate_configuration


@validate_configuration
def extract_and_aggregate(
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Extractions and aggregations started!", "start")

    storage.delete(StoragePath.extracted)
    AggregatedStorage.delete(storage)

    # retrieve configuration
    settings = SettingsStorage.read_from_storage(storage)
    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    sites = SiteStorage.read_from_storage(storage, settings)
    extractors = ExtractorStorage.read_from_storage(storage)
    extractors_instances = [ex.instanciate(settings) for ex in extractors]

    print_extractors(extractors)

    # build timelines
    timelines = create_timelines(
        sites=sites,
        integrations=integrations,
        storage=storage,
        settings=settings,
    )

    tw = TimelineWalker()
    tw.storage = storage
    tw.bands = bands
    tw.integrations = integrations
    tw.timelines = timelines
    tw.extractors = extractors_instances

    # walk intervals in timelines
    for (
        interval_data,
        labels,
        interval_details,
        interval,
        band,
        extractor,
        timeline,
    ) in tw.walk():
        if len(interval_data) == 0:
            continue

        # Aggregate
        aggregated_data = list(np.mean(interval_data, axis=0))

        AggregatedStorage.append_data(
            storage=storage,
            data=aggregated_data,
            band=band,
            integration=timeline.integration,
            extractor=extractor,
        )

        AggregatedStorage.append_site(
            storage=storage,
            site=timeline.site.name,
            band=band,
            integration=timeline.integration,
            extractor=extractor,
        )

        AggregatedStorage.append_interval_details(
            storage=storage,
            interval_details=interval_details,
            band=band,
            integration=timeline.integration,
            extractor=extractor,
        )

        AggregatedStorage.append_timestamp(
            storage=storage,
            timestamp=interval.start,
            band=band,
            integration=timeline.integration,
            extractor=extractor,
        )

        AggregatedStorage.append_labels(
            storage=storage,
            labels=labels,
            band=band,
            integration=timeline.integration,
            extractor=extractor,
        )

    # tw.print_leftovers()
    print_action("Extractions and aggregations completed!", "end")
    invoke_menu(storage, callback)



from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_no_reduced import print_no_reduced


def validate_reduced(action):
    """The reduced data validator."""

    def decorator(
        storage: Storage,
        callback: MenuCallback,
    ):
        if not storage.exists_dataset(StoragePath.reduced):
            print_no_reduced()
            invoke_menu(storage, callback)
            return

        action(storage, callback)

    return decorator



from InquirerPy.resolver import prompt
from rich import print

from processing.config.Config import Config


def ask_csv_path(config: Config) -> str:
    questions = [
        {
            "type": "input",
            "name": "filename",
            "message": "A .csv file will be created next to your storage file."
            " What should be its name?",
        },
    ]

    print()
    answers = prompt(questions)
    filename: str = str(answers["filename"])

    storage_filename = config.settings.storage_path.split("/")[-1]

    if filename == "":
        filename = storage_filename

    path = config.settings.storage_path.replace(storage_filename, f"{filename}.csv")
    return path



from typing import List

from processing.common.Timeline import Timeline
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.settings.SettingsConfig import SettingsConfig
from processing.config.sites.SiteConfig import SiteConfig
from processing.storage.Storage import Storage


def create_timelines(
    sites: List[SiteConfig],
    integrations: List[IntegrationConfig],
    settings: SettingsConfig,
    storage: Storage,
) -> List[Timeline]:
    timelines: List[Timeline] = []

    for site in sites:
        for integration in integrations:
            timeline = Timeline(
                site=site,
                integration=integration,
                storage=storage,
                origin=settings.timeline_origin,
                # debug=True,
            )

            timelines.append(timeline)

    return timelines



from datetime import datetime

from rich import print

from processing.common.Interval import Interval
from processing.common.Timeline import Timeline


def print_timeline_progress(
    timeline: Timeline,
    interval: Interval,
) -> None:
    start_seconds = interval.start / 1000
    date_format = "%Y-%m-%d %H:%M:%S"
    human_date = datetime.fromtimestamp(start_seconds).strftime(date_format)
    site_name = timeline.site.name
    integration_text = f"{timeline.integration.seconds}s"

    print(human_date, site_name, integration_text)



from typing import List

from processing.config.autoclusters.AutoclusterConfig import AutoclusterConfig
from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.labels.LabelConfig import LabelConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class AutoclusteredStorage:
    @staticmethod
    def delete(storage: Storage) -> None:
        storage.delete(StoragePath.autoclustered)

    @staticmethod
    def exists(storage: Storage) -> bool:
        return storage.exists_dataset(StoragePath.autoclustered)

    @staticmethod
    def read_from_storage(
        storage: Storage,
        autoclusters: List[AutoclusterConfig],
        band: BandConfig,
        integration: IntegrationConfig,
    ) -> List[LabelConfig]:
        autoclustereds: List[LabelConfig] = []

        for index, autocluster in enumerate(autoclusters):
            autoclustered = LabelConfig(index, autocluster.name)

            path = (
                f"{StoragePath.autoclustered.value}"
                f"/{band.name}"
                f"/{integration.seconds}"
                f"/{autocluster.index}"
            )

            values = storage.read(path)

            autoclustered.load_values(values[:])

            autoclustereds.append(autoclustered)

        return autoclustereds

    @staticmethod
    def get_path(
        band: BandConfig,
        integration: IntegrationConfig,
        ac: AutoclusterConfig,
    ) -> str:
        return (
            f"{StoragePath.autoclustered.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{ac.index}"
        )

    @staticmethod
    def write(
        storage: Storage,
        band: BandConfig,
        integration: IntegrationConfig,
        ac: AutoclusterConfig,
    ) -> None:
        path = AutoclusteredStorage.get_path(band, integration, ac)

        attributes = {
            "min_cluster_size": ac.min_cluster_size,
            "min_samples": ac.min_samples,
            "alpha": ac.alpha,
            "epsilon": ac.epsilon,
            "name": ac.name,
            "index": ac.index,
        }

        storage.write(
            path=path,
            data=ac.values,
            compression=True,
            attributes=attributes,
        )



from typing import List

from processing.config.extractors.ExtractorConfig import ExtractorConfig


def filter_nn_extractors(extractors: List[ExtractorConfig]) -> List[ExtractorConfig]:
    filtered_extractors: List[ExtractorConfig] = []

    for extractor in extractors:
        if extractor.name not in ExtractorConfig.nn_extractors:
            continue

        filtered_extractors.append(extractor)

    return filtered_extractors



from processing.common.AutoclusteredStorage import AutoclusteredStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_no_autoclustered import print_no_autoclustered


def validate_autoclustered(action):
    """The autoclusters generated data validator."""

    def decorator(
        storage: Storage,
        callback: MenuCallback,
    ):
        if not AutoclusteredStorage.exists(storage):
            print_no_autoclustered()
            invoke_menu(storage, callback)
            return

        action(storage, callback)

    return decorator



import re

from h5py import Group

from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.validate_configuration import validate_configuration

regex = "([0-9]{13})\/([0-9]{0,5})(.*)"


def fix_string(string: str) -> str:
    """
    Failing strings:
        1668596700000/0/home/user/samples/audio/.wav
        1668596700000/15000/home/user/samples/audio/.wav

        1668596700000/0Z:\sample\audio\.wav
        1668596700000/15000Z:\sample\audio\.wav

    Strings to have:
        1668596700000/0//home/user/samples/audio/.wav
        1668596700000/0/Z:\sample\audio\.wav
    """
    m = re.search(regex, string)
    assert m is not None, f"Unable to find match for {string}"
    return f"{m.group(0)}/{m.group(1)}/{m.group(2)}"


@validate_configuration
def fix_audio_windows_10_7_2(
    storage: Storage,
    callback: MenuCallback,
):
    """Fix audio on Windows happening in versions before 10.7.2"""

    print_action("Audio on Windows for versions <=10.7.2 started!", "start")

    path = f"{StoragePath.aggregated_interval_details.value}"
    group: Group = storage.read(path)

    for band in group.keys():
        band_path = f"{path}/{band}"
        band_group: Group = storage.read(band_path)

        for integration in band_group.keys():
            integration_path = f"{band_path}/{integration}"
            integration_group: Group = storage.read(integration_path)

            for dataset_index in integration_group.keys():
                dataset_path = f"{integration_path}/{dataset_index}"
                dataset = storage.read(dataset_path)

                for i in range(len(dataset)):
                    old_string = dataset.asstr()[i][0]
                    dataset[i] = [fix_string(old_string)]

    print_action("Audio on Windows for versions <=10.7.2 started!", "end")

    invoke_menu(storage, callback)



from processing.config.Config import Config
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_no_configuration import print_no_configuration


def validate_configuration_with_config(action):
    def decorator(
        config: Config,
        storage: Storage,
        callback: MenuCallback = None,
    ):
        if not Config.exists_in_storage(storage):
            print_no_configuration()
            invoke_menu(storage, callback)
            return

        return action(config, storage, callback)

    return decorator



from typing import List

from rich.console import Console
from rich.table import Table

from processing.config.digesters.DigesterConfig import DigesterConfig


def print_digesters(digesters: List[DigesterConfig]):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Name")
    table.add_column("Digester")

    for digester in digesters:
        table.add_row(
            digester.name,
            DigesterConfig.digesters[digester.name].__name__,
        )

    console.print(table)



from typing import List

import numpy as np

from processing.config.labels.LabelConfig import LabelConfig
from processing.digesters.Digester import Digester


class MeanSpreadingDigester(Digester):
    def digest(self, labels: List[LabelConfig]):
        label = labels[0]
        data = []

        for _, frame, _ in self.walk_within_label(label):
            percentile_95th = np.nanpercentile(frame, 95, axis=0)
            percentile_5th = np.nanpercentile(frame, 5, axis=0)
            mean = np.mean(percentile_95th - percentile_5th)
            data.append(float(mean))

        return data



import datetime
from typing import Union

from pandas import Timestamp


def convert_date_to_timestamp(date_string: Union[Timestamp, str]) -> int:
    """Converts a date to timestamp.

    The resulting timestamp is in UNIX format and milliseconds.

    Args:
        date: A pandas Timestamp object.

    Returns:
        The timestamp in milliseconds as an integer.
    """

    if isinstance(date_string, str):
        date = datetime.datetime.strptime(date_string, "%Y-%m-%d %H:%M:%S")
        return int(date.timestamp() * 1000)

    timestamp_seconds = datetime.datetime.timestamp(date_string)
    timestamp_milliseconds = timestamp_seconds * 1000

    return int(timestamp_milliseconds)



from abc import ABC, abstractmethod
from typing import List, Union

from h5py import Dataset


class AbstractReducer(ABC):
    _dimensions: int
    _seed: Union[None, int]
    _features: Union[None, Dataset]
    values: List[List[float]]

    def __init__(
        self,
    ) -> None:
        self._features = None

    @abstractmethod
    def calculate(self) -> List[List[float]]:
        pass

    def _validate_load(self) -> Dataset:
        if self._features is None:
            raise RuntimeError("Unable to find data for reducer. Please load first.")

        return self._features

    def load(
        self,
        dimensions: int,
        seed: Union[None, int],
        features: Dataset,
    ):
        self._dimensions = dimensions
        self._seed = seed
        self._features = features

    def set_values(
        self,
        features: List[List[float]],
    ) -> List[List[float]]:
        self.values = features
        return self.values



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple

from h5py import Dataset
from numpy import ndarray
from pandas import DataFrame, Index

from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.labels.LabelConfig import LabelConfig
from processing.storage.Storage import Storage

Features = List[List[float]]
Digested = List[Any]


class Digester(ABC):
    index: int
    is_pairing: bool
    __features: Optional[DataFrame] = None
    __labels: Optional[List[LabelConfig]] = None
    __storage: Optional[Storage] = None
    __band: Optional[BandConfig] = None
    __integration: Optional[IntegrationConfig] = None

    @property
    def features(self) -> DataFrame:
        assert self.__features is not None, "Please define features"
        return self.__features

    @features.setter
    def features(self, features: Dataset) -> None:
        self.__features = DataFrame(features)

    @property
    def labels(self) -> List[LabelConfig]:
        assert self.__labels is not None, "Please define labels"
        return self.__labels

    @labels.setter
    def labels(self, labels: List[LabelConfig]) -> None:
        self.__labels = labels

    @property
    def storage(self) -> Storage:
        assert self.__storage is not None, "Please define storage"
        return self.__storage

    @storage.setter
    def storage(self, storage: Storage) -> None:
        self.__storage = storage

    @property
    def band(self) -> BandConfig:
        assert self.__band is not None, "Please define band"
        return self.__band

    @band.setter
    def band(self, band: BandConfig) -> None:
        self.__band = band

    @property
    def integration(self) -> IntegrationConfig:
        assert self.__integration is not None, "Please define integration"
        return self.__integration

    @integration.setter
    def integration(self, integration: IntegrationConfig) -> None:
        self.__integration = integration

    def get_label_data(self, label: LabelConfig) -> Tuple[DataFrame, Index]:
        df = self.features
        df.index = label.values
        values: Index = df.index.unique()  # type: ignore

        return df, values

    def walk_within_label(self, label: LabelConfig):
        df, values = self.get_label_data(label)

        for index, value in enumerate(values):
            frame: ndarray = df[df.index == value].to_numpy()
            yield index, frame, value

    def walk_labels(self):
        if self.is_pairing:
            for label_a in self.labels:
                for label_b in self.labels:
                    data = self.digest([label_a, label_b])
                    yield data, label_a, label_b
        else:
            for label in self.labels:
                data = self.digest([label])
                yield data, label, None

    @abstractmethod
    def digest(self, labels: List[LabelConfig]) -> Digested:
        pass



from typing import List

import numpy as np
from h5py import Dataset

from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class MeanDistancesMatrix:
    @staticmethod
    def calculate(
        features: List[Dataset],
    ) -> List[List[float]]:
        from sklearn import metrics

        samples_count = features[0].shape[0]
        mean_distances_matrix = np.zeros([samples_count, samples_count])

        for i in range(len(features)):
            previous_mean_distances_matrix = mean_distances_matrix

            umap = features[i]

            current_mean_distances_matrix = metrics.pairwise_distances(umap)

            mean_distances_matrix = (
                (previous_mean_distances_matrix * i) + current_mean_distances_matrix
            ) / (i + 1)

        values = mean_distances_matrix.tolist()

        return values

    @staticmethod
    def read_from_storage(
        storage: Storage,
        band: BandConfig,
        integration: IntegrationConfig,
        trim_half: bool = False,
    ):
        path = MeanDistancesMatrix.get_path(band, integration)
        dataset = storage.read(path)
        matrix = np.array(dataset)

        if trim_half is True:
            np.fill_diagonal(matrix, 0)

        return matrix

    @staticmethod
    def get_path(
        band: BandConfig,
        integration: IntegrationConfig,
    ) -> str:
        return (
            f"{StoragePath.mean_distances_matrix.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
        )

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(StoragePath.mean_distances_matrix)



from typing import List

from rich.console import Console
from rich.table import Table

from processing.config.autoclusters.AutoclusterConfig import AutoclusterConfig


def print_autoclusters(autoclusters: List[AutoclusterConfig]):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Autocluster")
    table.add_column("Min cluster size")
    table.add_column("Min samples")
    table.add_column("Alpha")
    table.add_column("Epsilon")

    for autocluster in autoclusters:
        table.add_row(
            autocluster.name,
            str(autocluster.min_cluster_size),
            str(autocluster.min_samples),
            str(autocluster.alpha),
            str(autocluster.epsilon),
        )

    console.print(table)



from typing import List

import numpy as np

from processing.common.AutoclusteredStorage import AutoclusteredStorage
from processing.config.autoclusters.AutoclusterStorage import AutoclusterStorage
from processing.config.bands.BandConfig import BandConfig
from processing.config.extractors.ExtractorConfig import ExtractorConfig
from processing.config.files.FileStorage import FileStorage
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.labels.LabelConfig import LabelConfig
from processing.config.labels.LabelStorage import LabelStorage
from processing.config.settings.SettingsStorage import SettingsStorage
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class AggregatedLabelStorage:
    @staticmethod
    def read_from_storage(
        storage: Storage,
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: ExtractorConfig,
    ) -> List[LabelConfig]:
        aggregated_labels: List[LabelConfig] = []

        # Fetching autoclustered values first
        autoclusters = AutoclusterStorage.read_from_storage(storage)
        autoclustereds = AutoclusteredStorage.read_from_storage(
            storage=storage,
            autoclusters=autoclusters,
            band=band,
            integration=integration,
        )

        for autoclustered in autoclustereds:
            aggregated_labels.append(autoclustered)

        # Fetching aggregated labels (interval labels after walking timeline intervals)
        settings = SettingsStorage.read_from_storage(storage)
        files = FileStorage.read_from_storage(storage, settings)
        files_labels = LabelStorage.read_from_storage(storage, files)
        properties = [fl.property for fl in files_labels]

        path = (
            f"{StoragePath.aggregated_labels.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{extractor.index}"
        )

        aggregated_values = storage.read(path)

        # INFO: As labels can be empty, this check is needed
        if aggregated_values.shape[1] == 0:
            aggregated_values_strings = []
        else:
            aggregated_values_strings = aggregated_values.asstr()

        aggregated_values_np = np.array(aggregated_values_strings)

        for index, values in enumerate(aggregated_values_np.T):
            property_ = properties[index]
            aggregated_label = LabelConfig(
                index=index + len(autoclustereds),
                property=property_,
            )
            aggregated_label.load_values(values)
            aggregated_labels.append(aggregated_label)

        return aggregated_labels



from processing.common.ComputationUmapStorage import ComputationUmapStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_no_computation_umap import print_no_computation_umap


def validate_computation_umap(action):
    def decorator(
        storage: Storage,
        callback: MenuCallback,
    ):
        if not ComputationUmapStorage.exists_in_storage(storage):
            print_no_computation_umap()
            invoke_menu(storage, callback)
            return

        return action(storage, callback)

    return decorator



from typing import List

from rich.console import Console
from rich.table import Table

from processing.config.extractors.ExtractorConfig import ExtractorConfig


def print_extractors(extractors: List[ExtractorConfig]):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Name")
    table.add_column("Extractor")
    table.add_column("Offset (ms)")
    table.add_column("Step (ms)")
    table.add_column("Persist (extracted data)")

    for extractor in extractors:
        table.add_row(
            extractor.name,
            ExtractorConfig.extractors[extractor.name].__name__,
            str(extractor.offset),
            str(extractor.step),
            "[green]:heavy_check_mark:[/green]"
            if extractor.persist
            else "[red]:x:[/red]",
        )

    console.print(table)



from typing import List

from rich.console import Console
from rich.table import Table

from processing.config.labels.LabelConfig import LabelConfig


def print_labels(labels: List[LabelConfig]):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Index")
    table.add_column("Property")
    table.add_column("Set")
    table.add_column("Values count")

    for label in labels:
        table.add_row(
            str(label.index),
            label.property,
            ", ".join([unique for unique in label.uniques]),
            str(len(label.values)),
        )

    console.print(table)



from typing import Union


# Use for testing purposes to artificially limit the amount of memory used
def limit_memory_resource(
    gigabytes: Union[int, float],
) -> None:
    import resource

    memory_limit_in_bytes = int(1024 * 1024 * 1024 * gigabytes)
    limits = (memory_limit_in_bytes, memory_limit_in_bytes)
    resource.setrlimit(resource.RLIMIT_AS, limits)



from typing import List

from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.config.trajectories.TrajectoryConfig import TrajectoryConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class RelativeTracedStorage:
    @staticmethod
    def delete(storage: Storage) -> None:
        storage.delete(StoragePath.relative_traced)
        storage.delete(StoragePath.relative_traced_relative_timestamps)

    @staticmethod
    def exists(storage: Storage) -> bool:
        return storage.exists_dataset(
            StoragePath.relative_traced
        ) and storage.exists_dataset(StoragePath.relative_traced_relative_timestamps)

    @staticmethod
    def get_distance_path(
        trajectory: TrajectoryConfig,
        ar: AggregatedReduceable,
    ) -> str:
        return (
            f"{StoragePath.relative_traced.value}"
            f"/{ar.band.name}"
            f"/{ar.integration.seconds}"
            f"/{ar.extractor.index}"
            f"/{trajectory.index}"
        )

    @staticmethod
    def get_timestamp_path(
        trajectory: TrajectoryConfig,
        ar: AggregatedReduceable,
    ) -> str:
        return (
            f"{StoragePath.relative_traced_relative_timestamps.value}"
            f"/{ar.band.name}"
            f"/{ar.integration.seconds}"
            f"/{ar.extractor.index}"
            f"/{trajectory.index}"
        )

    @staticmethod
    def write_distances(
        storage: Storage,
        trajectory: TrajectoryConfig,
        ar: AggregatedReduceable,
        relative_distances: List[float],
        label_property: str,
        label_value: str,
    ) -> None:
        path = RelativeTracedStorage.get_distance_path(
            trajectory=trajectory,
            ar=ar,
        )

        storage.write(
            path=path,
            data=relative_distances,
            compression=True,
            attributes={
                "extractor_index": str(ar.extractor.index),
                "trajectory_index": str(trajectory.index),
                "trajectory_name": trajectory.name,
                "label_property": label_property,
                "label_value": label_value,
            },
        )

    @staticmethod
    def write_timestamps(
        storage: Storage,
        trajectory: TrajectoryConfig,
        ar: AggregatedReduceable,
        relative_timestamps: List[List[float]],
    ) -> None:
        path = RelativeTracedStorage.get_timestamp_path(
            trajectory=trajectory,
            ar=ar,
        )

        storage.write(
            path=path,
            data=relative_timestamps,
            compression=True,
            attributes={
                "extractor_index": str(ar.extractor.index),
                "trajectory_index": str(trajectory.index),
            },
        )



from typing import List

from rich.console import Console
from rich.table import Table

from processing.config.reducers.ReducerConfig import ReducerConfig


def print_reducers(reducers: List[ReducerConfig]):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Name")
    table.add_column("Reducer")
    table.add_column("Dimensions")
    table.add_column("Bands")
    table.add_column("Integrations")
    table.add_column("Ranges")

    for reducer in reducers:
        table.add_row(
            reducer.name,
            ReducerConfig.algorithms[reducer.name].__name__,
            str(reducer.dimensions),
            str(", ".join([b.name for b in reducer.bands])),
            str(", ".join([i.name for i in reducer.integrations])),
            str(
                "*"
                if len(reducer.ranges) == 0
                else ", ".join([r.name for r in reducer.ranges])
            ),
        )

    console.print(table)



from typing import List

from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.config.bands.BandStorage import BandStorage
from processing.config.extractors.ExtractorStorage import ExtractorStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.storage.Storage import Storage
from processing.utils.filter_nn_extractors import filter_nn_extractors


def build_aggregated_reduceables(
    storage: Storage,
) -> List[AggregatedReduceable]:
    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    extractors = ExtractorStorage.read_from_storage(storage)
    nn_extractors = filter_nn_extractors(extractors)

    aggregated_reduceables = AggregatedReduceable.reconstruct(
        bands=bands,
        integrations=integrations,
        nn_extractors=nn_extractors,
    )

    return aggregated_reduceables



from typing import Any, List

from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.extractors.Extractor import Extractor
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class AggregatedStorage:
    @staticmethod
    def delete(storage: Storage) -> None:
        storage.delete(StoragePath.aggregated)
        storage.delete(StoragePath.aggregated_sites)
        storage.delete(StoragePath.aggregated_interval_details)
        storage.delete(StoragePath.aggregated_timestamps)
        storage.delete(StoragePath.aggregated_labels)

    @staticmethod
    def exists(storage: Storage) -> bool:
        return (
            storage.exists_dataset(StoragePath.aggregated)
            and storage.exists_dataset(StoragePath.aggregated_sites)
            and storage.exists_dataset(StoragePath.aggregated_interval_details)
            and storage.exists_dataset(StoragePath.aggregated_timestamps)
            and storage.exists_dataset(StoragePath.aggregated_labels)
        )

    @staticmethod
    def get_data_path(
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> str:
        return (
            f"/{StoragePath.aggregated.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{extractor.index}"
        )

    @staticmethod
    def append_data(
        storage: Storage,
        data: List[Any],
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> None:
        path = AggregatedStorage.get_data_path(
            band=band,
            integration=integration,
            extractor=extractor,
        )

        storage.append(
            path=path,
            data=[data],
            compression=True,
            attributes={
                "extractor": extractor.__class__.__name__,
                "offset": str(extractor.offset),
                "step": str(extractor.step),
                "method": "for step in file in site",
            },
        )

    @staticmethod
    def get_site_path(
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> str:
        return (
            f"/{StoragePath.aggregated_sites.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{extractor.index}"
        )

    @staticmethod
    def append_site(
        storage: Storage,
        site: str,
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> None:
        path = AggregatedStorage.get_site_path(
            band=band,
            integration=integration,
            extractor=extractor,
        )

        storage.append(
            path=path,
            data=[[site]],
        )

    @staticmethod
    def get_interval_details_path(
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> str:
        return (
            f"/{StoragePath.aggregated_interval_details.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{extractor.index}"
        )

    @staticmethod
    def append_interval_details(
        storage: Storage,
        interval_details: List[str],
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> None:
        path = AggregatedStorage.get_interval_details_path(
            band=band,
            integration=integration,
            extractor=extractor,
        )

        storage.append(
            path=path,
            data=[interval_details],
            attributes={
                "description": "Each column is {block_start}"
                "/{relative_file_start}/{file_path}",
            },
        )

    @staticmethod
    def get_timestamp_path(
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> str:
        return (
            f"/{StoragePath.aggregated_timestamps.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{extractor.index}"
        )

    # INFO: This stores duplicated data as timestamps are the same for
    # band and extractor given a single integration
    @staticmethod
    def append_timestamp(
        storage: Storage,
        timestamp: int,
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> None:
        path = AggregatedStorage.get_timestamp_path(
            band=band,
            integration=integration,
            extractor=extractor,
        )

        storage.append(
            path=path,
            data=[[timestamp]],
        )

    @staticmethod
    def get_labels_path(
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> str:
        return (
            f"/{StoragePath.aggregated_labels.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{extractor.index}"
        )

    @staticmethod
    def append_labels(
        storage: Storage,
        labels: List[str],
        band: BandConfig,
        integration: IntegrationConfig,
        extractor: Extractor,
    ) -> None:
        path = AggregatedStorage.get_labels_path(
            band=band,
            integration=integration,
            extractor=extractor,
        )

        storage.append(
            path=path,
            data=[labels],
        )



import numpy as np
from h5py import Dataset, Group

from processing.config.Config import Config
from processing.config.bands.BandStorage import BandStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.ask_band import ask_band
from processing.utils.ask_integration import ask_integration
from processing.utils.ask_npy_path import ask_npy_path
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.validate_mean_distances_matrix_with_config import (
    validate_mean_distances_matrix_with_config,
)


@validate_mean_distances_matrix_with_config
def export_computation_umaps(
    config: Config,
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Computation UMAPs export started!", "start")

    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)

    band = ask_band(bands)
    integration = ask_integration(integrations)

    group_path = (
        f"{StoragePath.computation_umap.value}/{band.name}/{integration.seconds}"
    )
    group: Group = storage.read(group_path)  # type: ignore

    datasets = []
    length = group.__len__()

    for i in range(length):
        dataset: Dataset = group.get(f"{i}")  # type: ignore
        datasets.append(np.array(dataset[:]))

    datasets_np = np.array(datasets)

    npy_path = ask_npy_path(config)
    np.save(npy_path, datasets_np)

    print_action("Computation UMAPs export completed!", "end")
    invoke_menu(storage, callback)



from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class SoundscapeIndexExtractor(Extractor):
    """ndsi"""

    def extract(self, loader: Loader):
        from maad import features

        data = []

        for slice_ in self.sound_walk(loader):
            spectrogram = loader.sound.get_spectrogram(slice_)

            ndsi, _, _, _ = features.soundscape_index(
                Sxx_power=spectrogram.s,
                fn=spectrogram.fn,
            )

            data.append([ndsi])

        return data



from typing import List

from pandas import pandas

from processing.config.ConfigParser import ConfigParser
from processing.config.digesters.DigesterConfig import DigesterConfig
from processing.config.digesters.DigesterSheet import DigesterSheet
from processing.config.ExcelSheet import ExcelSheet
from processing.storage.Storage import Storage


class DigesterStorage:
    names = "/digesters/names"

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(DigesterStorage.names)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(DigesterStorage.names)

    @staticmethod
    def write_to_storage(digesters: List[DigesterConfig], storage: Storage) -> None:
        names = DigesterConfig.flatten(digesters)

        storage.write(path=DigesterStorage.names, data=names)

    @staticmethod
    def read_from_storage(storage: Storage) -> List[DigesterConfig]:
        names_dataset = storage.read(DigesterStorage.names)
        names = storage.convert_dataset_to_string_list(names_dataset)
        digests = DigesterConfig.reconstruct(names)
        return digests

    @staticmethod
    def read_from_config(parser: ConfigParser) -> List[DigesterConfig]:
        sheet = ExcelSheet.digesters

        df = pandas.DataFrame([parser.get(sheet, DigesterSheet.name_)])
        df = df.T
        df = df.dropna(how="all")
        df.columns = ["names"]

        names = df["names"].tolist()

        digests = DigesterConfig.reconstruct(names)
        return digests



from rich import print

from processing.constants import APP_NAME
from processing.utils.get_version_from_setup import get_version_from_setup


def print_welcome() -> None:
    version = get_version_from_setup()
    print(
        f"[green]"
        f":waving_hand: Welcome to [bold]{APP_NAME}[/bold]"
        f" "
        f"[italic]v{version}[/italic]"
        f"[/green]"
    )
    print()



from typing import List

from processing.common.ContinuousTimeTrajectory import ContinuousTimeTrajectory
from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.config.reducers.ReducerConfig import ReducerConfig


class TrajectoryConfig:
    step_by_name = {
        "hour": 60 * 60,
        "day": 60 * 60 * 24,
        "month": 60 * 60 * 24 * 30,
    }

    def __init__(
        self,
        index: int,
        name: str,
        start: int,
        end: int,
        label_property: str,
        label_value: str,
        step: int,  # seconds
    ) -> None:
        self.index = index
        self.name = name
        self.start = start
        self.end = end
        self.label_property = label_property
        self.label_value = label_value
        self.step = step

    @staticmethod
    def flatten(trajectories: List["TrajectoryConfig"]):
        names = [t.name for t in trajectories]
        starts = [t.start for t in trajectories]
        ends = [t.end for t in trajectories]
        label_properties = [t.label_property for t in trajectories]
        label_values = [t.label_value for t in trajectories]
        steps = [t.step for t in trajectories]

        return names, starts, ends, label_properties, label_values, steps

    @staticmethod
    def reconstruct(
        names: List[str],
        starts: List[int],
        ends: List[int],
        label_properties: List[str],
        label_values: List[str],
        steps: List[int],
    ) -> List["TrajectoryConfig"]:
        trajectories = []

        for index, name in enumerate(names):
            start = starts[index]
            end = ends[index]
            label_property = label_properties[index]
            label_value = label_values[index]
            step = steps[index]

            trajectory = TrajectoryConfig(
                index=index,
                name=name,
                start=start,
                end=end,
                label_property=label_property,
                label_value=label_value,
                step=step,
            )

            trajectories.append(trajectory)

        return trajectories

    def create_instance(
        self,
        band: BandConfig,
        integration: IntegrationConfig,
        reducer: ReducerConfig,
    ) -> ContinuousTimeTrajectory:
        self.band = band
        self.integration = integration
        self.reducer = reducer
        self.instance = ContinuousTimeTrajectory()
        return self.instance



from pandas import DataFrame

from processing.common.AggregatedLabelStorage import AggregatedLabelStorage
from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.config.bands.BandStorage import BandStorage
from processing.config.Config import Config
from processing.config.extractors.ExtractorStorage import ExtractorStorage
from processing.config.files.FileSheet import FileSheet
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.config.ranges.RangeStorage import RangeStorage
from processing.config.reducers.ReducerStorage import ReducerStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.ask_band import ask_band
from processing.utils.ask_csv_path import ask_csv_path
from processing.utils.ask_integration import ask_integration
from processing.utils.filter_nn_extractors import filter_nn_extractors
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.validate_configuration_with_config import (
    validate_configuration_with_config,
)


@validate_configuration_with_config
def export_dataframe(
    config: Config,
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Export started", "start")

    payload = {}

    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    extractors = ExtractorStorage.read_from_storage(storage)
    nn_extractors = filter_nn_extractors(extractors)

    band = ask_band(bands)
    integration = ask_integration(integrations)
    csv_path = ask_csv_path(config)

    timestamps_path = (
        f"{StoragePath.aggregated_timestamps.value}"
        f"/{band.name}"
        f"/{integration.seconds}"
        f"/{extractors[0].index}"
    )
    aggregated_timestamps = storage.read(timestamps_path)

    # Interval indexes
    payload["interval_index"] = []
    interval_index = 0
    for _ in aggregated_timestamps:
        payload["interval_index"].append(str(interval_index))
        interval_index += 1

    aggregated_sites_path = (
        f"{StoragePath.aggregated_sites.value}"
        f"/{band.name}"
        f"/{integration.seconds}"
        f"/{extractors[0].index}"
    )
    aggregated_sites = storage.read(aggregated_sites_path)
    payload["site"] = [s[0].decode("utf-8") for s in aggregated_sites]

    # Timestamps
    payload["timestamp"] = [at[0] for at in aggregated_timestamps]

    # Labels
    aggregated_labels = AggregatedLabelStorage.read_from_storage(
        storage=storage,
        band=band,
        integration=integration,
        extractor=extractors[0],
    )
    for al in aggregated_labels:
        key = f"{FileSheet.label_prefix.value}{al.property}"
        payload[key] = al.values

    # Reduced features
    ranges = RangeStorage.read_from_storage(storage)
    reducers = ReducerStorage.read_from_storage(storage, bands, integrations, ranges)

    aggregated_reduceables = AggregatedReduceable.reconstruct(
        bands=bands,
        integrations=integrations,
        nn_extractors=nn_extractors,
    )

    ars = list(
        filter(
            lambda ar_: ar_.band == band and ar_.integration == integration,
            aggregated_reduceables,
        )
    )

    for reducer in reducers:
        reducer.load(band, integration)

        if not reducer.should_calculate():
            continue

        for ar in ars:
            aggregated_features = storage.read(ar.get_reduced_path(reducer))
            for d in range(reducer.dimensions):
                key = f"{reducer.index}_{reducer.name}_{reducer.dimensions}d_{d+1}"
                payload[key] = [af[d] for af in aggregated_features]

    # Aggregated data
    for extractor in extractors:
        path = (
            f"{StoragePath.aggregated.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
            f"/{extractor.index}"
        )
        data = storage.read(path)
        for index in range(data.shape[1]):
            key = f"{extractor.index}_{extractor.name}_{index+1}"
            payload[key] = [d[index] for d in data]

    df = DataFrame(payload)
    df.to_csv(csv_path, index=False)

    print_action("Export completed!", "end")
    invoke_menu(storage, callback)



from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage


def invoke_menu(
    storage: Storage,
    callback: MenuCallback,
) -> None:
    if callback is None:
        return

    callback(storage)



from typing import List, Tuple


class IntegrationConfig:
    def __init__(
        self,
        index: int,
        name: str,
        duration: int,
    ) -> None:
        self.index: int = index
        self.name: str = name
        self.__duration: int = duration  # seconds

    @staticmethod
    def flatten(
        integrations: List["IntegrationConfig"],
    ) -> Tuple[List[str], List[int]]:
        names = [i.name for i in integrations]
        durations = [i.__duration for i in integrations]
        return names, durations

    @staticmethod
    def reconstruct(
        names: List[str],
        durations: List[int],
    ) -> List["IntegrationConfig"]:
        integrations = []

        for index, name in enumerate(names):
            integration = IntegrationConfig(
                index=index,
                name=name,
                duration=durations[index],
            )

            integrations.append(integration)

        return integrations

    @property
    def seconds(self) -> int:
        return self.__duration

    @property
    def milliseconds(self) -> int:
        return self.__duration * 1000



from rich.console import Console


def prettify_exceptions(action):
    def decorator(*args, **kwargs):
        try:
            return action(*args, **kwargs)
        except Exception:
            console = Console()
            console.print_exception(show_locals=True)

    return decorator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 11 10:46:51 2023

@author: rj102145
"""

import warnings

import numpy as np
import scipy.fftpack as fftpack
import scipy.signal as signal
import torch

eps = 1e-12


def torch_leq(waveform, sample_rate, dt=0.125, ref=1):
    """
    computes the Leq Tensor from a torch Tensor such as given by torchaudio.load
    waveform, sample_rate = torchaudio.load(file_path)

    Parameters
    ----------
    waveform : torchTensor[nb_channels, nb_samples_per_channels]
        raw wav samples
    sample_rate : int
        sample rate in Hz
    dt : float
        integration time for the Leq values in seconds

    Returns
    -------
    Leq torch.Tensor[nb_channels, nb_Leq _values]

    """
    dn = int(dt * sample_rate)
    n_ch, l_sig = waveform.shape
    n_leq = l_sig // dn
    if n_leq >= 1:
        waveform = waveform[:, : n_leq * dn]
        waveform = waveform.reshape([n_ch, n_leq, dn])
        short_time_leq = 10 * torch.log10(
            torch.nanmean(waveform**2, -1) / (ref**2) + eps
        )
    else:
        warnings.warn("Too few wav samples according to the Leq integration time")
        short_time_leq = torch.empty([n_ch, n_leq])
    return short_time_leq


def numpy_leq(
    waveform,
    sample_rate,
    dt=0.125,
    ref=1,
    freq_lim=[],  # INFO: This argument is mutable.
):
    """
    computes the Leq np.array from a np.array such as given by
    scipy.io.wavfile.read
    import scipy.io.wavfile as wavfile
    sample_rate, waveform = wavfile.read(file_path)

    Parameters
    ----------
    waveform : np.array[nb_samples_per_channels, nb_channels]
        raw wav samples
    sample_rate : int
        sample rate in Hz
    dt : float
        integration time for the Leq values in seconds

    Returns
    -------
    Leq np.array[nb_channels, nb_Leq _values]

    """
    if len(freq_lim) == 2:
        if freq_lim[0] > 0:
            ph_filtre = signal.butter(
                5, freq_lim[0], btype="highpass", output="sos", fs=sample_rate
            )
            waveform = signal.sosfiltfilt(
                ph_filtre, waveform, axis=-1, padtype="odd", padlen=None
            )
        if freq_lim[1] < sample_rate:
            pb_filtre = signal.butter(
                5, freq_lim[1], btype="lowpass", output="sos", fs=sample_rate
            )
            waveform = signal.sosfiltfilt(
                pb_filtre, waveform, axis=-1, padtype="odd", padlen=None
            )

    dn = int(dt * sample_rate)
    if len(waveform.shape) == 1:
        waveform = np.reshape(waveform, (len(waveform), 1))
    l_sig, n_ch = waveform.shape
    n_leq = l_sig // dn
    if n_leq >= 1:
        waveform = waveform[: n_leq * dn, :]
        waveform = waveform.reshape([n_leq, dn, n_ch])
        short_time_leq = 10 * np.log10(np.nanmean(waveform**2, 1) / (ref**2) + eps)
    else:
        warnings.warn("Too few wav samples according to the Leq integration time")
        short_time_leq = np.empty((0,) * 2)
    return short_time_leq


def energy_fract_from_spect(waveform, sample_rate, dt=1.0):
    """
    Parameters
    ----------
    waveform : np.array[nb_samples_per_channels, nb_channels]
        raw wav samples
    sample_rate : int
        sample rate in Hz
    dt : float
        integration time for the fourier mean transform

    Returns
    -------
    cumFourier : np.array[nb_of_frequency, nb_channels]
        the cumulative sum of the mean spectrum power over the frequency axis
    freq : np.array[nb_of_frequency]
        the corresponding frequency values

    """
    dnfft = int(dt * sample_rate)
    if len(waveform.shape) == 1:
        waveform = np.reshape(waveform, (len(waveform), 1))
    l_sig, n_ch = waveform.shape
    # number of zeros to pad
    if l_sig % dnfft == 0:
        n_pad = 0
    else:
        n_pad = dnfft - l_sig % dnfft
    # 'padding' coefficient
    alpha = dnfft / (dnfft - n_pad)
    # zero padding
    waveform = np.vstack((waveform, np.zeros((n_pad, n_ch))))
    nb_fft = waveform.shape[0] // dnfft
    waveform = waveform.reshape([dnfft, int(nb_fft), n_ch])
    fourierform = fftpack.fft(waveform, n=sample_rate, axis=0)
    # not considering the 0 Hz and (obviously) only the positives frequencies
    # energy per frequency steps
    # import matplotlib.pyplot as plt
    # plt.plot(10*np.log10(np.abs(fourierform)**2))
    fourierform = fourierform[1 : int(np.ceil((dnfft) / 2)), :, :]
    fourierform = np.abs(fourierform) ** 2
    fourierform = fourierform * alpha
    fourierform = np.nansum(fourierform, axis=1)
    # frequency steps
    freq = fftpack.fftfreq(dnfft, d=1 / sample_rate)
    freq = freq[1 : int(np.ceil((dnfft) / 2))]
    # cumulative energie per frequency steps
    cum_fourier = np.cumsum(fourierform, axis=0)

    return fourierform, cum_fourier, np.array(freq)


def statistical_leq(short_time_leq, stat=None):
    if stat == None:
        leq_stat = 10 * np.log10(np.mean(10 ** (short_time_leq / 10)))
    else:
        leq_stat = np.quantile(short_time_leq, 1 - stat / 100)
    return leq_stat


def statistical_power_spectrum(
    cum_fourier,
    freq,
    stat,
    channel=0,
    freq_lim=[],  # INFO: This argument is mutable.
):
    cum_fourier2 = cum_fourier[:, channel]
    if len(freq_lim) == 2:
        cum_fourier2 = cum_fourier2[(freq >= freq_lim[0]) * (freq <= freq_lim[1])]
        freq = freq[(freq >= freq_lim[0]) * (freq <= freq_lim[1])]
    spect_stat = cum_fourier2[-1] * (stat / 100)
    freq_stat = float(freq[cum_fourier2 <= spect_stat][-1])
    return freq_stat


def energy_ratio(fourierform, freq, freq_lim, channel=0, ratio_level=0.5):
    fourierform = fourierform[:, channel]
    center_freq = (freq_lim[1] - freq_lim[0]) * ratio_level + freq_lim[0]
    fourier_form_a = fourierform[(freq >= freq_lim[0]) * (freq <= center_freq)]
    fourier_form_b = fourierform[(freq >= center_freq) * (freq <= freq_lim[1])]
    nrj_ratio = 10 * np.log10(np.mean(fourier_form_a)) - 10 * np.log10(
        np.mean(fourier_form_b)
    )
    return nrj_ratio



from processing.utils.pack_trajectories import PackedTrajectories


def walk_packed_trajectories(packed_trajectories: PackedTrajectories):
    for label_property in packed_trajectories.keys():
        for label_value in packed_trajectories[label_property].keys():
            yield (
                label_property,
                label_value,
                packed_trajectories[label_property][label_value],
            )



from processing.config.files.FileConfig import FileConfig
from processing.loaders.SoundLoader import SoundLoader
from processing.loaders.TorchLoader import TorchLoader


class Loader:
    torch: TorchLoader
    sound: SoundLoader

    def __init__(self) -> None:
        self.torch = TorchLoader()
        self.sound = SoundLoader()

    def load(self, file: FileConfig):
        self.torch.load(file)
        self.sound.load(file)

    def release(self):
        self.torch.release()
        self.sound.release()

    def destroy(self):
        self.torch.destroy()
        self.sound.destroy()



from typing import List

from InquirerPy.resolver import prompt

from processing.config.bands.BandConfig import BandConfig


def ask_band(bands: List[BandConfig]) -> BandConfig:
    questions = [
        {
            "type": "list",
            "name": "choices",
            "choices": [band.name for band in bands],
            "message": "Choose your band",
        }
    ]

    answers = prompt(questions)
    name: str = str(answers["choices"])

    results: List[BandConfig] = list(filter(lambda b: b.name == name, bands))

    assert len(results) == 1, f"{len(results)} results for {name}"

    band = results[0]

    return band



from processing.config.files.FileConfig import FileConfig


class TimelineIntervalOverlapError(Exception):
    def __init__(self, new_file: FileConfig, existing_file: FileConfig) -> None:
        super().__init__(
            f"Unable to add duplicate file #{new_file.index}."
            f" Interval overlapping with file #{existing_file.index}."
        )



from processing.common.AutoclusteredStorage import AutoclusteredStorage
from processing.common.MeanDistancesMatrix import MeanDistancesMatrix
from processing.config.autoclusters.AutoclusterStorage import AutoclusterStorage
from processing.config.bands.BandStorage import BandStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.print_autoclusters import print_autoclusters
from processing.utils.validate_autoclusters import validate_autoclusters
from processing.utils.validate_configuration import validate_configuration
from processing.utils.validate_mean_distances_matrix import (
    validate_mean_distances_matrix,
)
from processing.utils.walk_bands_integrations import walk_bands_integrations


@validate_configuration
@validate_mean_distances_matrix
@validate_autoclusters
def autocluster(
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Autoclustering started!", "start")

    AutoclusteredStorage.delete(storage)

    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    autoclusters = AutoclusterStorage.read_from_storage(storage)
    print_autoclusters(autoclusters)

    for band, integration in walk_bands_integrations(bands, integrations):
        for ac in autoclusters:
            ac.create_instance(band, integration)
            mdm = storage.read(MeanDistancesMatrix.get_path(band, integration))
            ac.calculate(mdm)
            AutoclusteredStorage.write(storage, band, integration, ac)

    print_action("Autoclustering completed!", "end")
    invoke_menu(storage, callback)



from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class BioacousticsIndexExtractor(Extractor):
    """bi"""

    def extract(self, loader: Loader):
        from maad import features

        data = []

        for slice_ in self.sound_walk(loader):
            spectrogram = loader.sound.get_spectrogram(slice_, "amplitude")

            bi = features.bioacoustics_index(
                Sxx=spectrogram.s,
                fn=spectrogram.fn,
            )

            data.append([bi])

        return data



from rich import print


def print_new_line(is_replace: bool = False) -> None:
    string = "[bold]---[/bold]"

    if is_replace is True:
        print(string, end="\r")
        return

    print(string)



from typing import List

from rich.console import Console
from rich.table import Table

from processing.config.trajectories.TrajectoryConfig import TrajectoryConfig
from processing.utils.convert_timestamp_to_date import convert_timestamp_to_date


def print_trajectories(trajectories: List[TrajectoryConfig]):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Trajectory")
    table.add_column("Start")
    table.add_column("End")
    table.add_column("Label property")
    table.add_column("Label value")
    table.add_column("Step")

    for trajectory in trajectories:
        table.add_row(
            trajectory.name,
            convert_timestamp_to_date(trajectory.start),
            convert_timestamp_to_date(trajectory.end),
            trajectory.label_property,
            trajectory.label_value,
            str(trajectory.step),
        )

    console.print(table)



import torch
from torch import Tensor, nn


class VGG(nn.Module):
    def __init__(self, features):
        super(VGG, self).__init__()
        self.features = features
        self.embeddings = nn.Sequential(
            nn.Linear(512 * 4 * 6, 4096), nn.ReLU(True),
            nn.Linear(4096, 4096), nn.ReLU(True),
            nn.Linear(4096, 128), nn.ReLU(True)
        )

    def forward(self, x: Tensor):
        with torch.no_grad():
            x = self.features(x)
            # Transpose the output from features to
            # remain compatible with VGGish embeddings
            x = torch.transpose(x, 1, 3)
            x = torch.transpose(x, 1, 2)
            x = x.contiguous()
            x = x.view(x.size(0), -1)
            return self.embeddings(x)



from typing import List

import numpy as np

from processing.digesters.Digester import Digester


class OverlapDigester(Digester):
    def digest(self, labels):
        label = labels[0]
        dataframe, values = self.get_label_data(label)

        nb_clusters = len(values)

        plus_limits = []
        minus_limits = []

        for value in values:
            plus_limits.append(
                np.nanpercentile(
                    dataframe[dataframe.index == value].to_numpy(),
                    95,
                    axis=0,
                )
            )
            minus_limits.append(
                np.nanpercentile(
                    dataframe[dataframe.index == value].to_numpy(),
                    5,
                    axis=0,
                )
            )

        dataframe_b = np.zeros([nb_clusters, nb_clusters])

        for i, value in enumerate(values):
            dataframe_a = dataframe[dataframe.index == value].to_numpy()

            mask_a = (dataframe_a > minus_limits[i]) * (dataframe_a < plus_limits[i])

            dim_to_keep = np.where(np.sum(mask_a, axis=0) == 0, np.nan, 1)

            for j in range(nb_clusters):
                mask_b = (dataframe_a > minus_limits[j]) * (
                    dataframe_a < plus_limits[j]
                )

                dataframe_b[i, j] = np.nanmean(
                    np.sum(mask_a * mask_b * dim_to_keep, axis=0)
                    / np.sum(mask_a, axis=0)
                )

        overlaps: List[List[float]] = list((dataframe_b + dataframe_b.T) / 2)

        return overlaps



from typing import Dict, List, Optional

from rich import print
from rich.progress import track

from processing.common.FileLoader import FileLoader
from processing.common.Interval import Block, Interval
from processing.common.Timeline import Timeline
from processing.config.bands.BandConfig import BandConfig
from processing.config.files.FileConfig import FileConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.extractors.Extractor import Extracted, Extractor
from processing.storage.Storage import Storage
from processing.utils.print_new_line import print_new_line
from processing.utils.print_timeline_progress import print_timeline_progress

FileIndex = int
BandIndex = int
ExtractorIndex = int
BlockIndex = int


class TimelineWalker:
    """The walker for timelines

    Feed multiple timelines to enumerate their intervals.

    The name can be misleading as this class takes multiple timelines as input.
    This behaviour allows keeping track of loaded files and extracted data (in
    order to do it once per file and extractor).
    """

    def __init__(self) -> None:
        self.__timelines: Optional[List[Timeline]] = None
        self.__bands: Optional[List[BandConfig]] = None
        self.__integrations: Optional[List[IntegrationConfig]] = None
        self.__extractors: Optional[List[Extractor]] = None
        self.__storage: Optional[Storage] = None

        self.loaders: Dict[FileIndex, FileLoader] = {}
        self.extracted: Dict[
            FileIndex, Dict[BandIndex, Dict[ExtractorIndex, Extracted]]
        ] = {}

        # Current file index
        self.cfi = None

    @property
    def timelines(self) -> List[Timeline]:
        assert self.__timelines is not None, "Please attach timelines"
        return self.__timelines

    @timelines.setter
    def timelines(self, timelines: List[Timeline]) -> List[Timeline]:
        self.__timelines = timelines
        return self.__timelines

    @property
    def bands(self) -> List[BandConfig]:
        assert self.__bands is not None, "Please attach bands"
        return self.__bands

    @bands.setter
    def bands(self, bands: List[BandConfig]) -> List[BandConfig]:
        self.__bands = bands
        return self.__bands

    @property
    def integrations(self) -> List[IntegrationConfig]:
        assert self.__integrations is not None, "Please attach integrations"
        return self.__integrations

    @integrations.setter
    def integrations(
        self, integrations: List[IntegrationConfig]
    ) -> List[IntegrationConfig]:
        self.__integrations = integrations
        return self.__integrations

    @property
    def extractors(self) -> List[Extractor]:
        assert self.__extractors is not None, "Please attach extractors"
        return self.__extractors

    @extractors.setter
    def extractors(self, extractors: List[Extractor]) -> List[Extractor]:
        self.__extractors = extractors
        return self.__extractors

    @property
    def storage(self) -> Storage:
        assert self.__storage is not None, "Please attach storage"
        return self.__storage

    @storage.setter
    def storage(self, storage: Storage) -> Storage:
        self.__storage = storage
        return self.__storage

    def print_leftovers(self):
        print_new_line()
        print("Memory leftovers")
        # show extracted data
        for k, v in self.extracted.items():
            for kk, vv in v.items():
                for kkk, vvv in vv.items():
                    print(k, kk, kkk, len(vvv[0]))

        # show loaded files
        for _, loader in self.loaders.items():
            print(type(loader.loader.torch.audio))
            loader.loader.sound.print_leftovers()

    def __enumerate(self):
        for timeline in track(self.timelines, description=("Extracting...")):
            for interval in timeline.intervals:
                print_timeline_progress(timeline, interval)
                for band_index, band in enumerate(self.bands):
                    for extractor in self.extractors:
                        yield (
                            timeline,
                            interval,
                            band_index,
                            band,
                            extractor,
                        )

    def load_file(self, file: FileConfig) -> FileLoader:
        if file.index in self.loaders.keys():
            return self.loaders[file.index]

        loader = FileLoader(file)
        loader.load()
        self.loaders[file.index] = loader
        return self.loaders[file.index]

    def prepare_extracted(self, file: FileConfig) -> None:
        if file.index in self.extracted.keys():
            return

        self.extracted[file.index] = {}

    def load_band(self, band: BandConfig, file: FileConfig):
        if band.index in self.extracted[file.index].keys():
            return

        self.extracted[file.index][band.index] = {}

    def run_extraction(
        self,
        extractor: Extractor,
        band: BandConfig,
        file: FileConfig,
        timeline: Timeline,
    ):
        if extractor.index in self.extracted[file.index][band.index].keys():
            return

        extractor.band = band

        loader = self.loaders[file.index]

        self.extracted[file.index][band.index][extractor.index] = extractor.extract(
            loader.loader
        )

        if extractor.is_persist:
            # Store only for first integration
            # this could be above
            if timeline.integration.index == 0:
                extractor.store(
                    self.extracted[file.index][band.index][extractor.index],
                    self.storage,
                )

    def get_block_data(
        self,
        block: Block,
        band: BandConfig,
        extractor: Extractor,
    ):
        offset = extractor.offset / 1000  # seconds floats
        step = extractor.step / 1000  # seconds float

        start = (block.start - block.file.start) // 1000  # seconds
        duration = (block.end - block.start) // 1000  # seconds

        start = int((start + offset) / step)
        duration = int(duration / step)

        end = start + duration

        file_data = self.extracted[block.file.index][band.index][extractor.index]

        block_data = file_data[start:end]

        return block_data

    def get_block_details(
        self,
        block: Block,
    ) -> str:
        file_relative_start = block.start - block.file.start
        block_details = f"{block.start}/{file_relative_start}/{block.file.path}"
        return block_details

    def purge(
        self,
        f: FileIndex,
        b: BandIndex,
        extractor: Extractor,
        b_: BlockIndex,
        interval: Interval,
        timeline: Timeline,
    ):
        # Removing extracted data
        # INFO: this lefts over the last operation in memory
        if self.cfi is not None and f != self.cfi:
            del self.extracted[self.cfi]

        # Unloading file
        is_last_integration = timeline.integration.index == len(self.integrations) - 1
        is_last_band = b == len(self.bands) - 1
        is_last_extractor = extractor.index == len(self.extractors) - 1
        is_last_block = b_ == len(interval.blocks) - 1

        is_last_op = (
            is_last_integration and is_last_band and is_last_extractor and is_last_block
        )

        if self.cfi is not None and is_last_op:
            self.loaders[self.cfi].release()

    def walk(self):
        for timeline, interval, b, band, extractor in self.__enumerate():
            interval_data = []
            labels: List[str] = []
            interval_details: List[str] = []

            for b_, block in enumerate(interval.blocks):
                f = block.file.index

                self.load_file(file=block.file)
                self.prepare_extracted(file=block.file)
                self.load_band(band=band, file=block.file)

                self.run_extraction(
                    extractor=extractor,
                    band=band,
                    file=block.file,
                    timeline=timeline,
                )

                block_data = self.get_block_data(
                    block=block,
                    band=band,
                    extractor=extractor,
                )

                interval_data = [*interval_data, *block_data]
                labels = [*labels, *block.file.labels]

                block_details = self.get_block_details(block=block)
                if block_details not in interval_details:
                    interval_details.append(block_details)

                self.purge(f, b, extractor, b_, interval, timeline)

                self.cfi = f

            yield (
                interval_data,
                labels,
                interval_details,
                interval,
                band,
                extractor,
                timeline,
            )



from typing import Any, List


def validate_excel_names(names: List[str | Any]) -> List[str]:
    names = [str(n) for n in names]
    assert len(names) == len(set(names)), "Names must be unique"
    return names



from enum import Enum


class StoragePath(Enum):
    """The paths for all data located in the storage file.

    File type:
        HDF

    Details:
        https://www.wikiwand.com/en/Hierarchical_Data_Format
    """

    #################
    # Configuration #
    #################

    settings = "/settings"
    """The information gathered from the configuration file.

    This will contain multiple groups and datasets corresponding to each tab
    of the user Excel configuration file.

    User will need to refresh using `pnpm process:config` to update data located
    inside the storage file.

    Type:
        Group
    """

    #######################
    # Configuration/Bands #
    #######################

    # INFO: This allows duplicates that can cause collisions because bands
    # are referenced by their names and not absolute indexes.
    # A solution would be to reference by index or force user to have unique names.
    bands_names = "/bands/names"
    """The names of specified bands.

    Each row contains a string.

    Type:
        Dataset
    """

    bands_lows = "/bands/lows"
    """The low frequencies for each band.

    Each row contains one integer for the frequency in Hertz.

    Type:
        Dataset
    """

    bands_highs = "/bands/highs"
    """The high frequencies for each band.

    Each row contains one integer for the frequency in Hertz.

    Type:
        Dataset
    """

    ##############################
    # Configuration/Integrations #
    ##############################

    integrations_names = "/integrations/names"
    """The names of specified integrations.

    Each row contains a string.

    Type:
        Dataset
    """

    integrations_seconds = "/integrations/seconds"
    """The duration in seconds for each band.

    Each row contains an integer.

    Type:
        Dataset
    """

    ########################
    # Configuration/Ranges #
    ########################

    # INFO: This allows duplicates thus can cause collisions.
    # See `Configuration/Bands` annotation for more details.
    ranges_names = "/ranges/names"
    """The names of specified ranges.

    Each row contains a string.

    Type:
        Dataset
    """

    ranges_starts = "/ranges/starts"
    """The start date for each range.

    Each row contains a timestamp as an integer.

    Timestamps are in UNIX format and milliseconds.

    Type:
        Dataset
    """

    ranges_ends = "/ranges/ends"
    """The end date for each range.

    Each row contains a timestamp as an integer.

    Timestamps are in UNIX format and milliseconds.

    Type:
        Dataset
    """

    #########
    # Files #
    #########

    files_names = "/files/names"
    """The file paths relative to the audio folder.

    Each row contains a string.

    Each string starts with a `/`.

    This `/` is a convention because it gets merged with the audio folder path.

    Type:
        Dataset
    """

    files_sites = "/files/sites"
    """The site for each file.

    Each row contains a string.

    Type:
        Dataset
    """

    files_labels = "/files/labels"
    """The labels for each file.

    Each row contains `n` columns of strings.

    Each column has the index of the corresponding label property.

    Type:
        Dataset
    """

    files_timestamps = "/files/timestamps"
    """The timestamp for each file.

    Timestamps are in UNIX format and milliseconds.
    Each row contains an integer.

    Type:
        Dataset
    """

    files_durations = "/files/durations"
    """The audio durations for each file.

    Audio durations are in seconds.
    Each row contains an integer.

    Type:
        Dataset
    """

    ######################
    # Configuration/Meta #
    ######################

    labels_properties = "/labels/properties"
    """The meta properties specified in the configuration file.

    Each row contains a string.

    Type:
        Dataset
    """

    labels_sets = "/labels/sets"
    """The meta unique values (sets).

    Each row contains a variable number of strings (separated as columns).

    Each string represent a possible value given
    the meta property (referenced by its index).

    Meta sets are filled with empty strings `""` to keep the dataset rectangular.

    Type:
        Dataset
    """

    ##############################
    # Configuration/Autoclusters #
    ##############################

    autoclusters_names = "/autoclusters/names"
    """The names of autoclusters requested in the configuration file.

    Each row is a string referring to a particular algorithm.

    Enumeration:
        ../clusterings/ClusteringName.py

    Type:
        Dataset
    """

    autoclusters_min_cluster_sizes = "/autoclusters/min_cluster_sizes"
    """The minimum count of elements within a given cluster.

    Each row contains an integer.

    Type:
        Dataset
    """

    autoclusters_min_samples = "/autoclusters/min_samples"
    """The minimum samples for autoclustering algorithms.

    Each row contains an integer.

    Type:
        Dataset
    """

    autoclusters_alphas = "/autoclusters/alphas"
    """The alpha values for autoclustering algorithms.

    Each row contains a float.

    Type:
        Dataset
    """

    autoclusters_epsilons = "/autoclusters/epsilons"
    """The epsilon values for autoclustering algorithms.

    Each row contains a float.

    Type:
        Dataset
    """

    ##############################
    # Configuration/Trajectories #
    ##############################

    trajectories_names = "/trajectories/names"
    """The names of trajectories requested in the configuration file.

    Each row is a string.

    Type:
        Dataset
    """

    trajectories_starts = "/trajectories/starts"
    """The start date for each trajectory.

    Each row contains a timestamp as an integer.

    Timestamps are in UNIX format and milliseconds.

    Type:
        Dataset
    """

    trajectories_ends = "/trajectories/ends"
    """The end date for each trajectory.

    Each row contains a timestamp as an integer.

    Timestamps are in UNIX format and milliseconds.

    Type:
        Dataset
    """

    trajectories_label_properties = "/trajectories/label_properties"
    trajectories_label_values = "/trajectories/label_values"
    trajectories_steps = "/trajectories/steps"

    ##########################
    # Configuration/Reducers #
    ##########################

    reducers_names = "/reducers/names"
    """The names of reducers requested in the configuration file.

    Each row contains a string.

    Enumeration:
        ../reducers/ReducerName.py

    Factory:
        ../reducers/Reducer.py

    Type:
        Dataset
    """

    reducers_dimensions = "/reducers/dimensions"
    """The dimensions for each reducer.

    Each row contains an integer.

    Type:
        Dataset
    """

    reducers_bands = "/reducers/bands"
    """The bands for each reducer.

    Each row contains an array of strings (stored as columns) corresponding
    to the name of the band.

    Rows are filled with empty strings `""` if needed to keep the dataset rectangular.

    Type:
        Dataset
    """

    reducers_integrations = "/reducers/integrations"
    """The integration names for each reducer.

    Each row contains an array of strings (stored as columns) corresponding
    to the name of the integration.

    Rows are filled with empty strings `""` if needed to keep the dataset rectangular.

    Type:
        Dataset
    """

    reducers_ranges = "/reducers/ranges"
    """The range names for each reducer.

    Each row contains an array of strings (stored as columns) corresponding
    to the name of the range.

    Rows are filled with empty strings `""` if needed to keep the dataset rectangular.

    Type:
        Dataset
    """

    ##########################
    # Configuration/Pairings #
    ##########################

    pairings_names = "/pairings/names"
    """The pairing names requested in the configuration file.

    Each row contains a string.

    Enumeration:
        ../pairings/PairingName.py

    Factory:
        ../pairings/Pairing.py

    Type:
        Dataset
    """

    #####################
    # Computation UMAPs #
    #####################

    computation_umap = "/computation_umap"
    """The computation UMAPs needed for further data generation.

    Computation UMAPs are used for:
        - Mean distances matrix
        - Trajectories

    The number of computation UMAPs is specified in the configuration file.

    The group contains one dataset per computation UMAP index (iteration),
    band and integration value.

    Each dataset is named after the integration value.

    Each row contains an array (columns) of floats.

    This path is to be used as a prefix!

    Example:
        /computation_umap_{computation_umap_index}/{band}/{integration}

    Type:
        Group prefix
    """

    #########################
    # Mean distances matrix #
    #########################

    mean_distances_matrix = "/mean_distances_matrix"
    """The mean distances matrix for further data generation.

    Mean distances matrices are used for:
        - Autoclustering

    The group contains one dataset per band and integration value.
    Each dataset is named after the integration value.
    Each row contains an array (columns) of floats.

    Example:
        /mean_distances_matrix/{band}/{integration}

    Type:
        Group
    """

    #############
    # Data/Site #
    #############

    sites_names = "/sites/names"
    """The unique site names found in the project.

    Each row contains a string.

    Type:
        Dataset
    """

    # TODO: Store paths instead
    sites_file_indexes = "/sites/file_indexes"
    """The file indexes for each site.

    Each row contains one or more integers representing file indexes.

    The number of file indexes per site can varry.
    Making rectangular for storage and trimming for reconstructing is therefore needed.

    Type:
        Dataset
    """

    # Extractors
    extractors_names = "/extractors/names"
    extractors_offsets = "/extractors/offsets"
    extractors_steps = "/extractors/steps"
    extractors_persists = "/extractors/persists"
    extracted = "/extracted"

    aggregated = "/aggregated"
    aggregated_sites = "/aggregated_sites"
    aggregated_interval_details = "/aggregated_interval_details"
    aggregated_timestamps = "/aggregated_timestamps"
    aggregated_labels = "/aggregated_labels"
    reduced = "/reduced"
    autoclustered = "/autoclustered"
    traced = "/traced"
    traced_timestamps = "/traced_timestamps"
    traced_relative_timestamps = "/traced_relative_timestamps"
    digested = "/digested"
    config_file = "/config_file"
    relative_traced = "/relative_traced"
    relative_traced_relative_timestamps = "/relative_traced_relative_timestamps"



from processing.extractors.Extractor import Extractor
from processing.loaders.Loader import Loader


class LeqMaadExtractor(Extractor):
    def extract(self, loader: Loader):
        from maad import features

        sample_rate = loader.sound.sample_rate

        data = []

        for slice_ in self.sound_walk(loader):
            leq = features.temporal_leq(
                s=slice_.sound,
                fs=sample_rate,
                gain=42,
            )

            data.append([leq])

        return data



from signal import SIGINT, signal
from typing import Optional

from processing.actions.autocluster import autocluster
from processing.actions.compute_requirements import compute_requirements
from processing.actions.digest import digest
from processing.actions.export_computation_umaps import export_computation_umaps
from processing.actions.export_dataframe import export_dataframe
from processing.actions.export_mdm import export_mdm
from processing.actions.extract_and_aggregate import extract_and_aggregate
from processing.actions.fix_audio_windows_10_7_2 import fix_audio_windows_10_7_2
from processing.actions.purge_requirements import purge_requirements
from processing.actions.reduce import reduce
from processing.actions.refresh_configuration import refresh_configuration
from processing.actions.repack_storage import repack_storage
from processing.actions.trace_relative_trajectories import trace_relative_trajectories
from processing.actions.trace_trajectories import trace_trajectories
from processing.config.Config import Config
from processing.storage.Storage import Storage
from processing.utils.ask_menu import MenuChoice, ask_menu
from processing.utils.prettify_exceptions import prettify_exceptions
from processing.utils.print_welcome import print_welcome
from processing.utils.quit_sse import quit_sse

stored_config: Config
last_choice: Optional[MenuChoice] = None


@prettify_exceptions
def main(
    loaded_storage: Optional[Storage] = None,
    config_path: Optional[str] = None,
):
    """CLI entry point"""

    global stored_config
    global last_choice

    if config_path is not None:
        config = Config(path=config_path)
        stored_config = config
    else:
        config = stored_config

    if loaded_storage is None:
        print_welcome()
        storage = Storage(config.settings.storage_path)
    else:
        storage = loaded_storage

    signal(SIGINT, lambda _signum, _frame: quit_sse(storage))

    answer = ask_menu(storage, last_choice)
    last_choice = answer

    if answer == MenuChoice.RefreshConfig.value:
        refresh_configuration(config, storage, main)
    if answer == MenuChoice.ExtractAggregate.value:
        refresh_configuration(config, storage)
        extract_and_aggregate(storage, main)
    if answer == MenuChoice.Reduce.value:
        refresh_configuration(config, storage)
        reduce(storage, main)
    if answer == MenuChoice.ComputeRequirements.value:
        refresh_configuration(config, storage)
        compute_requirements(storage, main)
    if answer == MenuChoice.PurgeRequirements.value:
        refresh_configuration(config, storage)
        purge_requirements(storage, main)
    if answer == MenuChoice.Autocluster.value:
        refresh_configuration(config, storage)
        autocluster(storage, main)
    if answer == MenuChoice.Trace.value:
        refresh_configuration(config, storage)
        trace_trajectories(storage, main)
    if answer == MenuChoice.RelativeTrace.value:
        refresh_configuration(config, storage)
        trace_relative_trajectories(storage, main)
    if answer == MenuChoice.Digest.value:
        refresh_configuration(config, storage)
        digest(storage, main)
    if answer == MenuChoice.RunAll.value:
        refresh_configuration(config, storage)
        extract_and_aggregate(storage)
        reduce(storage)
        compute_requirements(storage)
        autocluster(storage)
        trace_trajectories(storage)
        trace_relative_trajectories(storage)
        digest(storage, main)
    if answer == MenuChoice.ExportDataframe.value:
        export_dataframe(config, storage, main)
    if answer == MenuChoice.ExportComputationUmaps.value:
        export_computation_umaps(config, storage, main)
    if answer == MenuChoice.ExportMeanDistancesMatrix.value:
        export_mdm(config, storage, main)
    if answer == MenuChoice.Repack.value:
        repack_storage(storage, main)
    if answer == MenuChoice.FixAudioWindows10_7_2.value:
        refresh_configuration(config, storage)
        fix_audio_windows_10_7_2(storage, main)
    else:
        quit_sse(storage)



from processing.config.ConfigParser import ConfigParser
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class BinaryStorage:
    config_file = StoragePath.config_file.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(BinaryStorage.config_file)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(BinaryStorage.config_file)

    @staticmethod
    def read_from_storage(storage: Storage) -> None:
        pass

    @staticmethod
    def read_from_config(parser: ConfigParser) -> None:
        pass

    @staticmethod
    def write_to_storage(path: str, storage: Storage) -> None:
        with open(path, "rb") as f:
            binary_data = f.read()

            storage.write_binary(
                path=BinaryStorage.config_file,
                binary_data=binary_data,
            )



from rich.console import Console
from rich.table import Table

from processing.config.settings.SettingsConfig import SettingsConfig
from processing.config.sites.SiteStorage import SiteStorage
from processing.storage.Storage import Storage


def print_file_indexes_by_site(storage: Storage, settings: SettingsConfig):
    console = Console()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Site")
    table.add_column("File indexes")

    sites = SiteStorage.read_from_storage(storage, settings)

    for site in sites:
        table.add_row(
            site.name,
            str(", ".join([str(f.index) for f in site.files])),
        )

    console.print(table)



from rich.progress import track

from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action


def purge_requirements(
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Requirements computation purge started!", "start")

    for _ in track(range(1)):
        storage.delete(StoragePath.computation_umap)
        storage.delete(StoragePath.mean_distances_matrix)

    print_action("Requirements computation purge completed!", "end")
    invoke_menu(storage, callback)



from typing import List

from pandas import DataFrame

from processing.constants import STRING_YES


def validate_excel_booleans(df: DataFrame, default: bool) -> List[bool]:
    return [True if p == STRING_YES else default for p in df.tolist()]



import subprocess

from rich import print

from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.ask_for_repack_replacement import ask_for_repack_replacement
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action


def repack_storage(
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Repacking started!", "start")

    # Need to close the storage first
    storage.close()

    path_repack = storage.path.replace(".h5", ".repack.h5")

    subprocess.run(
        [
            "h5repack",
            storage.path,
            f"{path_repack}",
        ]
    )

    print(f"Repacked storage path: {path_repack}")

    answer = ask_for_repack_replacement()

    print_action("Repacking completed!", "end")

    if answer is True:
        path_old = storage.path
        subprocess.run(["rm", path_old])
        subprocess.run(["mv", path_repack, path_old])

        storage_repacked = Storage(path_old)
        invoke_menu(storage_repacked, callback)
    else:
        storage.load()
        invoke_menu(storage, callback)



from typing import List


def reverse_array(array: List[List[str]]) -> List[List[str]]:
    reversed_array = list(zip(*array))
    reversed_array = list(list(sublist) for sublist in reversed_array)
    return reversed_array



import numpy as np

from processing.digesters.Digester import Digester


class SumStandardDeviationDigester(Digester):
    def digest(self, labels):
        label = labels[0]
        data = []

        for _, frame, _ in self.walk_within_label(label):
            std = np.std(frame, axis=0)
            sum_std = np.sum(std)
            data.append(float(sum_std))

        return data



from typing import List

from pandas import Timestamp

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.labels.LabelConfig import LabelConfig
from processing.config.trajectories.TrajectoryConfig import TrajectoryConfig
from processing.config.trajectories.TrajectoryExcel import TrajectoryExcel
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.convert_date_to_timestamp import convert_date_to_timestamp
from processing.utils.validate_excel_names import validate_excel_names
from processing.utils.validate_trajectory_labels import validate_trajectory_labels


class TrajectoryStorage:
    names = StoragePath.trajectories_names.value
    starts = StoragePath.trajectories_starts.value
    ends = StoragePath.trajectories_ends.value
    label_properties = StoragePath.trajectories_label_properties.value
    label_values = StoragePath.trajectories_label_values.value
    steps = StoragePath.trajectories_steps.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(TrajectoryStorage.names)
        storage.delete(TrajectoryStorage.starts)
        storage.delete(TrajectoryStorage.ends)
        storage.delete(TrajectoryStorage.label_properties)
        storage.delete(TrajectoryStorage.label_values)
        storage.delete(TrajectoryStorage.steps)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            storage.exists_dataset(TrajectoryStorage.names)
            and storage.exists_dataset(TrajectoryStorage.starts)
            and storage.exists_dataset(TrajectoryStorage.ends)
            and storage.exists_dataset(TrajectoryStorage.label_properties)
            and storage.exists_dataset(TrajectoryStorage.label_values)
            and storage.exists_dataset(TrajectoryStorage.steps)
        )

    @staticmethod
    def read_from_storage(storage: Storage) -> List[TrajectoryConfig]:
        names_dataset = storage.read(TrajectoryStorage.names)
        names = storage.convert_dataset_to_string_list(names_dataset)
        starts = storage.read(TrajectoryStorage.starts)
        ends = storage.read(TrajectoryStorage.ends)
        label_properties_dataset = storage.read(TrajectoryStorage.label_properties)
        label_properties = storage.convert_dataset_to_string_list(
            label_properties_dataset
        )
        label_values_dataset = storage.read(TrajectoryStorage.label_values)
        label_values = storage.convert_dataset_to_string_list(label_values_dataset)
        steps = storage.read(TrajectoryStorage.steps)

        trajectories = TrajectoryConfig.reconstruct(
            names=names,
            starts=starts[:],
            ends=ends[:],
            label_properties=label_properties,
            label_values=label_values,
            steps=steps[:],
        )

        return trajectories

    @staticmethod
    def read_from_config(
        parser: ConfigParser,
        labels: List[LabelConfig],
    ) -> List[TrajectoryConfig]:
        sheet = ExcelSheet.trajectories

        names = parser.get(sheet, TrajectoryExcel.name_)
        names = validate_excel_names(names)

        starts: List[Timestamp] = parser.get(sheet, TrajectoryExcel.start)
        ends: List[Timestamp] = parser.get(sheet, TrajectoryExcel.end)

        starts_timestamps = [convert_date_to_timestamp(start) for start in starts]
        ends_timestamps = [convert_date_to_timestamp(end) for end in ends]

        label_properties = parser.get(sheet, TrajectoryExcel.label_property)
        label_values = parser.get(sheet, TrajectoryExcel.label_value)
        step_names = parser.get(sheet, TrajectoryExcel.step)
        steps = [TrajectoryConfig.step_by_name[sn] for sn in step_names]

        validate_trajectory_labels(names, label_properties, label_values, labels)

        trajectories = TrajectoryConfig.reconstruct(
            names=names,
            starts=starts_timestamps,
            ends=ends_timestamps,
            label_properties=label_properties,
            label_values=label_values,
            steps=steps,
        )

        return trajectories

    @staticmethod
    def write_to_storage(
        trajectories: List[TrajectoryConfig],
        storage: Storage,
    ) -> None:
        (
            names,
            starts,
            ends,
            label_properties,
            label_values,
            steps,
        ) = TrajectoryConfig.flatten(trajectories)

        storage.write(path=TrajectoryStorage.names, data=names)
        storage.write(path=TrajectoryStorage.starts, data=starts)
        storage.write(path=TrajectoryStorage.ends, data=ends)
        storage.write(path=TrajectoryStorage.label_properties, data=label_properties)
        storage.write(path=TrajectoryStorage.label_values, data=label_values)
        storage.write(path=TrajectoryStorage.steps, data=steps)



from typing import List, Tuple

from processing.utils.get_file_full_path import get_file_full_path
from processing.utils.print_action import print_action


class FileConfig:
    index: int
    name: str
    timestamp: int  # UNIX timestamp in milliseconds
    site: str
    labels: List[str]
    duration: int  # milliseconds

    def __init__(
        self,
        index: int,
        name: str,
        timestamp: int,
        site: str,
        labels: List[str],
        duration: int,
        audio_path: str,
    ) -> None:
        self.index = index
        self.name = name
        self.timestamp = timestamp
        self.site = site
        self.labels = labels
        self.duration = duration
        self.__warn_small_duration(self.duration)
        self.audio_path = audio_path
        self.path = get_file_full_path(name, audio_path)

    @staticmethod
    def flatten(
        files: List["FileConfig"],
    ) -> Tuple[List[str], List[int], List[str], List[List[str]], List[int]]:
        names = [f.name for f in files]
        timestamps = [f.timestamp for f in files]
        sites = [f.site for f in files]
        labels = [f.labels for f in files]
        durations = [f.duration for f in files]

        return names, timestamps, sites, labels, durations

    @staticmethod
    def reconstruct(
        names: List[str],
        timestamps: List[int],
        sites: List[str],
        labels: List[List[str]],
        durations: List[int],
        audio_path: str,
    ) -> List["FileConfig"]:
        files = []

        for index, name in enumerate(names):
            label = FileConfig.extract_labels(labels, index)

            file_ = FileConfig(
                index=index,
                name=name,
                timestamp=timestamps[index],
                site=sites[index],
                labels=label,
                duration=durations[index],
                audio_path=audio_path,
            )

            files.append(file_)

        return files

    @property
    def seconds(self) -> int:
        return self.duration // 1000

    @property
    def milliseconds(self) -> int:
        return self.duration

    @property
    def start(self) -> int:
        return self.timestamp

    @property
    def end(self) -> int:
        return self.timestamp + self.duration

    def get_interval_count(self, interval_step: int) -> int:
        return (self.end - self.start) // interval_step

    @staticmethod
    def extract_labels(labels: List[List[str]], index: int) -> List[str]:
        try:
            label = labels[index]
        except IndexError:
            label = []

        return label

    def __warn_small_duration(self, duration: int) -> None:
        if duration < 1000:
            print_action(
                f"File index {self.index} has a duration of {duration} milliseconds.",
                "warning",
            )



from typing import List

from pandas import pandas

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.extractors.ExtractorConfig import ExtractorConfig
from processing.config.extractors.ExtractorDefaults import ExtractorDefaults
from processing.config.extractors.ExtractorSheet import ExtractorSheet
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.validate_excel_booleans import validate_excel_booleans
from processing.utils.validate_excel_ints import validate_excel_ints


class ExtractorStorage:
    names = StoragePath.extractors_names.value
    offsets = StoragePath.extractors_offsets.value
    steps = StoragePath.extractors_steps.value
    persists = StoragePath.extractors_persists.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(ExtractorStorage.names)
        storage.delete(ExtractorStorage.offsets)
        storage.delete(ExtractorStorage.steps)
        storage.delete(ExtractorStorage.persists)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            storage.exists_dataset(ExtractorStorage.names)
            and storage.exists_dataset(ExtractorStorage.offsets)
            and storage.exists_dataset(ExtractorStorage.steps)
            and storage.exists_dataset(ExtractorStorage.persists)
        )

    @staticmethod
    def read_from_storage(storage: Storage) -> List[ExtractorConfig]:
        names_dataset = storage.read(ExtractorStorage.names)
        names = storage.convert_dataset_to_string_list(names_dataset)

        offsets = storage.read(ExtractorStorage.offsets)
        steps = storage.read(ExtractorStorage.steps)
        persists_ints = storage.read(ExtractorStorage.persists)
        persists = [True if p == 1 else False for p in persists_ints]

        extractors = ExtractorConfig.reconstruct(
            names=names,
            offsets=offsets[:],
            steps=steps[:],
            persists=persists[:],
        )

        return extractors

    @staticmethod
    def read_from_config(parser: ConfigParser) -> List[ExtractorConfig]:
        sheet = ExcelSheet.extractors

        df = pandas.DataFrame(
            [
                parser.get(sheet, ExtractorSheet.name_),
                parser.get(sheet, ExtractorSheet.offset),
                parser.get(sheet, ExtractorSheet.step),
                parser.get(sheet, ExtractorSheet.persist),
            ]
        )
        df = df.T
        df = df.dropna(how="all")
        df.columns = ["names", "offsets", "steps", "persists"]

        names = df["names"].tolist()
        offsets = validate_excel_ints(df["offsets"], ExtractorDefaults.offset)
        steps = validate_excel_ints(df["steps"], ExtractorDefaults.step)
        persists = validate_excel_booleans(df["persists"], ExtractorDefaults.persist)

        extractors = ExtractorConfig.reconstruct(
            names=names,
            offsets=offsets,
            steps=steps,
            persists=persists,
        )

        return extractors

    @staticmethod
    def write_to_storage(extractors: List[ExtractorConfig], storage: Storage) -> None:
        names, offsets, steps, persists = ExtractorConfig.flatten(extractors)
        storage.write(ExtractorStorage.names, names)
        storage.write(ExtractorStorage.offsets, offsets)
        storage.write(ExtractorStorage.steps, steps)
        storage.write(ExtractorStorage.persists, persists)



from typing import Optional, Tuple

from processing.config.files.FileConfig import FileConfig


class TorchLoader:
    def __init__(self) -> None:
        from torch import Tensor

        self.audio: Optional[Tensor] = None
        self.sample_rate: Optional[int] = None

    def destroy(self):
        # this will raise `AttributeError` if you try loading the file again
        del self.audio
        del self.sample_rate

    def release(self):
        self.audio = None
        self.sample_rate = None

    def load(self, file: FileConfig):
        import torchaudio
        from torch import Tensor

        if self.audio is not None and self.sample_rate is not None:
            return self.audio, self.sample_rate

        loaded: Tuple[Tensor, int] = torchaudio.load(file.path)  # type: ignore
        self.audio, self.sample_rate = loaded
        return self.audio, self.sample_rate



from typing import List

from h5py import Dataset, Group

from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class ComputationUmapStorage:
    @staticmethod
    def delete(storage: Storage) -> None:
        storage.delete(StoragePath.computation_umap)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return storage.exists_dataset(StoragePath.computation_umap)

    @staticmethod
    def get_path(
        ar: AggregatedReduceable,
        index: int,  # Computation Index
    ) -> str:
        return (
            f"{StoragePath.computation_umap.value}"
            f"/{ar.band.name}"
            f"/{ar.integration.seconds}"
            f"/{index}"
        )

    @staticmethod
    def write(
        storage: Storage,
        ar: AggregatedReduceable,
        data: List[List[float]],
        index: int,
    ):
        path = ComputationUmapStorage.get_path(
            ar=ar,
            index=index,
        )

        storage.write(
            path=path,
            data=data,
            compression=True,
        )

    @staticmethod
    def read_from_storage(
        storage: Storage,
        band: BandConfig,
        integration: IntegrationConfig,
    ) -> List[Dataset]:
        path = (
            f"{StoragePath.computation_umap.value}"
            f"/{band.name}"
            f"/{integration.seconds}"
        )

        group: Group = storage.read(path)  # type: ignore

        datasets: List[Dataset] = []

        for dataset in group.values():
            datasets.append(dataset)

        return datasets



from dataclasses import dataclass
from typing import Union


@dataclass(frozen=True)
class SettingsConfig:
    storage_path: str
    audio_path: str
    expected_sample_rate: int
    timeline_origin: int

    # Optional
    audio_host: str
    timezone: Union[str, None]
    computation_umap_dimensions: int
    computation_umap_iterations: int
    display_umap_seed: int



from enum import Enum


class TrajectoryExcel(Enum):
    name_ = "trajectory"
    start = "start"
    end = "end"
    label_property = "label_property"
    label_value = "label_value"
    step = "step"



from dataclasses import dataclass, field
from typing import List

from processing.config.files.FileConfig import FileConfig


@dataclass(frozen=True, order=True)
class Block:  # aka Slice / EnrichedAudio
    start: int
    end: int
    file: FileConfig


@dataclass(frozen=True, order=True)
class Interval:
    start: int
    end: int
    blocks: List[Block] = field(default_factory=list)



import numpy as np
from pandas import DataFrame, Index
from sklearn import metrics

from processing.digesters.Digester import Digester


class DistanceDigester(Digester):
    def get_centers(self, dataframe: DataFrame, values: Index):
        # medioids
        centers = np.empty((len(values), dataframe.shape[1]))

        for index, _ in enumerate(values):
            centers[index, :] = np.nanpercentile(dataframe, 50, axis=0)

        return centers

    def digest(self, labels):
        label = labels[0]
        dataframe, values = self.get_label_data(label)
        centers = self.get_centers(dataframe, values)
        distances = metrics.pairwise_distances(centers, metric="manhattan")
        return distances



from typing import List

from processing.config.bands.BandConfig import BandConfig
from processing.config.bands.BandExcel import BandExcel
from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.validate_excel_names import validate_excel_names


class BandStorage:
    names = StoragePath.bands_names.value
    lows = StoragePath.bands_lows.value
    highs = StoragePath.bands_highs.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(BandStorage.names)
        storage.delete(BandStorage.lows)
        storage.delete(BandStorage.highs)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            storage.exists_dataset(BandStorage.names)
            and storage.exists_dataset(BandStorage.lows)
            and storage.exists_dataset(BandStorage.highs)
        )

    @staticmethod
    def read_from_storage(storage: Storage) -> List[BandConfig]:
        names_dataset = storage.read(BandStorage.names)

        names = storage.convert_dataset_to_string_list(names_dataset)
        lows = storage.read(BandStorage.lows)[:]
        highs = storage.read(BandStorage.highs)[:]

        bands = BandConfig.reconstruct(
            names=names,
            lows=lows,
            highs=highs,
        )

        return bands

    @staticmethod
    def write_to_storage(bands: List[BandConfig], storage: Storage) -> None:
        names, lows, highs = BandConfig.flatten(bands)

        storage.write(path=BandStorage.names, data=names)
        storage.write(path=BandStorage.lows, data=lows)
        storage.write(path=BandStorage.highs, data=highs)

    @staticmethod
    def read_from_config(parser: ConfigParser) -> List[BandConfig]:
        sheet = ExcelSheet.bands

        names: List[str] = parser.get(sheet, BandExcel.name_)
        names = validate_excel_names(names)

        lows: List[int] = parser.get(sheet, BandExcel.low)
        highs: List[int] = parser.get(sheet, BandExcel.high)

        bands = BandConfig.reconstruct(
            names=names,
            lows=lows,
            highs=highs,
        )

        return bands



from typing import List


def split_array_by_length(
    array: List[List[float]],
    length: int,
) -> List[List[List[float]]]:
    split_array = []

    elements_per_sublist = len(array) // length

    for _ in range(length):
        split_array.append([])

    for sublist_index in range(length):
        start = elements_per_sublist * sublist_index
        end = elements_per_sublist * (sublist_index + 1)

        sublist = array[start:end]

        for elements in sublist:
            split_array[sublist_index].append(elements)

    return split_array



from typing import List

import scipy
from numpy import ndarray
from pandas import DataFrame


def compute_relative_distances(
    path: DataFrame,
    starting_point: ndarray,
    mean_distance: float,
) -> List[List[float]]:
    relative_distance = (
        scipy.spatial.distance.cdist(
            path,
            starting_point,
            metric="euclidean",
        )
        / mean_distance
    )

    return relative_distance[:, 0]



from InquirerPy.base import Choice

from processing.common.AggregatedStorage import AggregatedStorage
from processing.common.AutoclusteredStorage import AutoclusteredStorage
from processing.common.MenuChoice import MenuChoice
from processing.common.RelativeTracedStorage import RelativeTracedStorage
from processing.common.TracedStorage import TracedStorage
from processing.config.Config import Config
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


def wrap_menu_choice_with_state(
    menu_choice: MenuChoice,
    storage: Storage,
) -> Choice:
    state = False

    if menu_choice is MenuChoice.RefreshConfig:
        state = Config.exists_in_storage(storage)
    elif menu_choice is MenuChoice.ExtractAggregate:
        state = AggregatedStorage.exists(storage)
    elif menu_choice is MenuChoice.Reduce:
        state = storage.exists_dataset(StoragePath.reduced)
    elif menu_choice is MenuChoice.ComputeRequirements:
        state = storage.exists_dataset(
            StoragePath.computation_umap
        ) and storage.exists_dataset(StoragePath.mean_distances_matrix)
    elif menu_choice is MenuChoice.Autocluster:
        state = AutoclusteredStorage.exists(storage)
    elif menu_choice is MenuChoice.Trace:
        state = TracedStorage.exists(storage)
    elif menu_choice is MenuChoice.RelativeTrace:
        state = RelativeTracedStorage.exists(storage)
    elif menu_choice is MenuChoice.Digest:
        state = storage.exists_dataset(StoragePath.digested)

    if state:
        return Choice(value=menu_choice.value, name=f"✅ {menu_choice.value}")
    else:
        return Choice(value=menu_choice.value, name=f"❌ {menu_choice.value}")



from typing import List

import numpy
from h5py import Dataset
from pandas import pandas

from processing.utils.filter_features_by_label_and_time import (
    filter_features_by_label_and_time,
)


class ContinuousTimeTrajectory:
    def __init__(self) -> None:
        self.__features = None
        self.__timestamps = None
        self.__timestamp_start = None
        self.__timestamp_end = None
        self.__labels_properties = None
        self.__labels_values = None
        self.__step = None

    def _validate_load(self):
        if (
            self.__features is None
            or self.__timestamps is None
            or self.__timestamp_start is None
            or self.__timestamp_end is None
            or self.__labels_properties is None
            or self.__labels_values is None
            or self.__step is None
        ):
            raise RuntimeError("Unable to find data for trajectory. Please load first.")

        return (
            self.__features,
            self.__timestamps,
            self.__timestamp_start,
            self.__timestamp_end,
            self.__labels_properties,
            self.__labels_values,
            self.__step,
        )

    def _set_values(self, values: List[List[float]]) -> List[List[float]]:
        self.values = values
        return self.values

    def set_timestamps(self, timestamps: List[int]) -> List[int]:
        self.timestamps = timestamps
        return self.timestamps

    def set_relative_timestamps(
        self,
        relative_timestamps: List[float],
    ) -> List[int]:
        self.relative_timestamps = relative_timestamps
        return self.relative_timestamps

    def load(
        self,
        features: Dataset,
        timestamps: Dataset,
        timestamp_start: int,
        timestamp_end: int,
        labels_properties: List[str],
        labels_values: Dataset,
        step: int,
    ):
        self.__features = features
        self.__timestamps = timestamps
        self.__timestamp_start = timestamp_start
        self.__timestamp_end = timestamp_end
        self.__labels_properties = labels_properties
        self.__labels_values = labels_values
        self.__step = step

    def calculate(
        self,
        trajectory_label_property: str,
        trajectory_label_value: str,
    ) -> List[List[float]]:
        (
            features,
            timestamps,
            timestamp_start,
            timestamp_end,
            labels_properties,
            labels_values,
            rolling_step,
        ) = self._validate_load()

        step = 1

        filtered_features, filtered_timestamps = filter_features_by_label_and_time(
            features=features,
            timestamps=timestamps,
            timestamp_start=timestamp_start,
            timestamp_end=timestamp_end,
            labels_properties=labels_properties,
            labels_values=labels_values,
            trajectory_label_property=trajectory_label_property,
            trajectory_label_value=trajectory_label_value,
        )

        self.set_timestamps(filtered_timestamps)

        dimensions = len(filtered_features[0])

        df_features = pandas.DataFrame(filtered_features)
        df_features.columns = [d for d in range(dimensions)]

        df_timestamps = pandas.DataFrame({"timestamps": filtered_timestamps})

        df = pandas.concat((df_features, df_timestamps), axis=1)

        df["relative_timestamps"] = df["timestamps"] - min(df["timestamps"])
        df["relative_timestamps"] = df["relative_timestamps"].apply(
            lambda ms: (ms / 1000) / rolling_step
        )

        rolling_count = len(df[df["relative_timestamps"] <= step])
        self.set_relative_timestamps(df["relative_timestamps"].tolist())

        # Calculating coordinates
        coordinates = numpy.empty([dimensions, len(df)])

        for d in range(dimensions):
            coordinates[d, :] = (
                df[d]
                .rolling(rolling_count, min_periods=1)
                .apply(lambda x: numpy.nanmean(x))
            )

        self._set_values(coordinates.T.tolist())
        return self.values



import argparse
import os
import platform
import sys

from processing.main import main
from processing.utils.append_to_config import append_to_config
from processing.utils.extract_config_from_storage import extract_config_from_storage
from processing.utils.prettify_exceptions import prettify_exceptions
from processing.utils.read_audio_path_from_config import read_audio_path_from_config
from processing.utils.walk_directory import walk_directory


def update_python_path():
    current_path = os.getcwd()

    if platform.system() == "Windows":
        processing_path = f"{current_path}\\processing"
    else:
        processing_path = f"{current_path}/processing"

    # Append `processing` path to PYTHONPATH
    if processing_path not in sys.path:
        sys.path.append(processing_path)


def parse_arguments() -> str:
    parser = argparse.ArgumentParser()
    parser.add_argument("config_path")
    args = parser.parse_args()
    return args.config_path


def start_processing():
    update_python_path()
    config_path = parse_arguments()
    main(config_path=config_path)


@prettify_exceptions
def extract_config():
    update_python_path()
    storage_path = parse_arguments()
    extract_config_from_storage(storage_path)


@prettify_exceptions
def fill_config():
    update_python_path()
    config_path = parse_arguments()
    audio_path = read_audio_path_from_config(config_path)
    paths = walk_directory(audio_path)
    append_to_config(config_path, paths)


def download():
    """Download dictionaries for offline use"""

    from rich import print
    from torch import hub

    hub.load_state_dict_from_url(
        "https://github.com/harritaylor/torchvggish/"
        "releases/download/v0.1/vggish-10086976.pth",
        progress=True,
    )

    print("You are now ready for offline use!")



from typing import List, Tuple

from processing.config.files.FileConfig import FileConfig


class SiteConfig:
    index: int
    name: str
    files: List[FileConfig]

    def __init__(
        self,
        index: int,
        name: str,
        files: List[FileConfig],
    ) -> None:
        self.index = index
        self.name = name
        self.files = files

    @staticmethod
    def flatten(sites: List["SiteConfig"]) -> Tuple[List[str], List[List[int]]]:
        names = [s.name for s in sites]
        file_indexes = [[f.index for f in s.files] for s in sites]

        return names, file_indexes

    @staticmethod
    def reconstruct(
        names: List[str],
        file_indexes: List[List[int]],
        files: List[FileConfig],
    ) -> List["SiteConfig"]:
        sites: List[SiteConfig] = []

        for index, name in enumerate(names):
            files_by_site: List[FileConfig] = []

            for file in files:
                if file.index in file_indexes[index]:
                    files_by_site.append(file)

            site: SiteConfig = SiteConfig(
                index=index,
                name=name,
                files=files_by_site,
            )

            sites.append(site)

        return sites



from typing import Dict, List, Tuple

from h5py import Dataset
from rich import print

from processing.config.bands.BandConfig import BandConfig
from processing.config.integrations.IntegrationConfig import IntegrationConfig
from processing.utils.print_action import print_action


class AutoclusterConfig:
    methods_by_name: Dict[str, str] = {
        "hdbscan-eom": "eom",
        "hdbscan-leaf": "leaf",
    }

    none_string = "None"

    def __init__(
        self,
        index: int,
        name: str,
        min_cluster_size: int,
        min_samples: str,
        alpha: float,
        epsilon: float,
    ) -> None:
        AutoclusterConfig.validate_name(name)

        self.index = index
        self.name = name
        self.min_cluster_size = min_cluster_size
        self.min_samples = min_samples
        self.alpha = float(alpha)
        self.epsilon = float(epsilon)

    @staticmethod
    def validate_name(name: str) -> None:
        """The validator for autoclustering names.

        Args:
            name: The autoclustering name to validate.

        Returns:
            None

        Raises:
            KeyError: An error occured because the autoclustering name
            has not been found.
        """
        if name in AutoclusterConfig.methods_by_name.keys():
            return

        raise KeyError(f"Unable to validate autoclustering name {name}.")

    @staticmethod
    def flatten(
        autoclusters: List["AutoclusterConfig"],
    ) -> Tuple[List[str], List[int], List[str], List[float], List[float]]:
        names = [ac.name for ac in autoclusters]
        min_cluster_sizes = [ac.min_cluster_size for ac in autoclusters]
        min_samples = [ac.min_samples for ac in autoclusters]
        alphas = [ac.alpha for ac in autoclusters]
        epsilons = [ac.epsilon for ac in autoclusters]

        return names, min_cluster_sizes, min_samples, alphas, epsilons

    @staticmethod
    def reconstruct(
        names: List[str],
        min_cluster_sizes: List[int],
        min_samples: List[str],
        alphas: List[float],
        epsilons: List[float],
    ) -> List["AutoclusterConfig"]:
        autoclusters = []

        for index, name in enumerate(names):
            autocluster = AutoclusterConfig(
                index=index,
                name=name,
                min_cluster_size=min_cluster_sizes[index],
                min_samples=min_samples[index],
                alpha=alphas[index],
                epsilon=epsilons[index],
            )

            autoclusters.append(autocluster)

        return autoclusters

    def create_instance(
        self,
        band: BandConfig,
        integration: IntegrationConfig,
    ):  # -> HDBSCAN
        from hdbscan import HDBSCAN

        self.band = band
        self.integration = integration

        method = self.methods_by_name[self.name]
        min_samples = (
            None if self.min_samples == self.none_string else int(self.min_samples)
        )

        self.instance = HDBSCAN(
            min_cluster_size=self.min_cluster_size,
            min_samples=min_samples,
            alpha=self.alpha,
            cluster_selection_epsilon=self.epsilon,
            cluster_selection_method=method,
            metric="precomputed",
            p=None,
            algorithm="best",
            leaf_size=50,
            approx_min_span_tree=True,
            gen_min_span_tree=False,
            core_dist_n_jobs=-1,
            match_reference_implementation=False,
        )

        return self.instance

    def calculate(
        self,
        mean_distances_matrix: Dataset,
    ) -> List[str]:
        try:
            clustering = self.instance.fit(mean_distances_matrix[:])
            labels = clustering.labels_.tolist()
        except MemoryError:
            print_action(
                f"The mean distances matrix {mean_distances_matrix.shape} "
                f"exceeds the available RAM.",
                "warning",
            )
            print("Filling with dummy results to continue processing...")

            labels = ["MemoryError"] * mean_distances_matrix.shape[0]

        print(set(labels))
        self.values = labels
        return self.values



from processing.config.autoclusters.AutoclusterStorage import AutoclusterStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_no_autoclusters import print_no_autoclusters


def validate_autoclusters(action):
    """The autoclusters configuration objects validator."""

    def decorator(
        storage: Storage,
        callback: MenuCallback,
    ):
        autoclusters = AutoclusterStorage.read_from_storage(storage)

        if len(autoclusters) == 0:
            print_no_autoclusters()
            invoke_menu(storage, callback)
            return

        action(storage, callback)

    return decorator



from enum import Enum


class SettingsRow(Enum):
    # Mandatory settings
    storage_path = "storage_path"
    audio_path = "audio_path"
    audio_host = "audio_host"
    expected_sample_rate = "expected_sample_rate"
    timeline_origin = "timeline_origin"

    # Optional settings
    timezone = "timezone"
    c_umap_dimensions = "computation_umap_dimensions"
    c_umap_iterations = "computation_umap_iterations"
    display_umap_seed = "display_umap_seed"



import numpy
import pandas as pd
from sklearn import metrics

from processing.digesters.Digester import Digester


class ContingencyDigester(Digester):
    def digest(self, labels):
        label_a = labels[0]
        label_b = labels[1]

        clusters_a = label_a.values
        clusters_b = label_b.values

        contingency_matrix = metrics.cluster.contingency_matrix(
            clusters_a,
            clusters_b,
        )

        # we don't need computing both 'a vs b' and 'b vs a' as they are
        # mirrored

        pairing_a = (
            contingency_matrix
            / numpy.tile(
                numpy.sum(contingency_matrix, axis=1),  # type: ignore
                (contingency_matrix.shape[1], 1),
            ).T
            * 100
        )

        payload_a = pd.DataFrame(pairing_a.T)

        return payload_a



from typing import List

from pandas import Timestamp

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.ranges.RangeConfig import RangeConfig
from processing.config.ranges.RangeExcel import RangeExcel
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.convert_date_to_timestamp import convert_date_to_timestamp
from processing.utils.validate_excel_names import validate_excel_names


class RangeStorage:
    names = StoragePath.ranges_names.value
    starts = StoragePath.ranges_starts.value
    ends = StoragePath.ranges_ends.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(RangeStorage.names)
        storage.delete(RangeStorage.starts)
        storage.delete(RangeStorage.ends)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            storage.exists_dataset(RangeStorage.names)
            and storage.exists_dataset(RangeStorage.starts)
            and storage.exists_dataset(RangeStorage.ends)
        )

    @staticmethod
    def read_from_storage(storage: Storage) -> List[RangeConfig]:
        names_dataset = storage.read(RangeStorage.names)

        names = storage.convert_dataset_to_string_list(names_dataset)
        starts = storage.read(RangeStorage.starts)
        ends = storage.read(RangeStorage.ends)

        ranges = RangeConfig.reconstruct(
            names=names,
            starts=starts[:],
            ends=ends[:],
        )

        return ranges

    @staticmethod
    def write_to_storage(ranges: List[RangeConfig], storage: Storage) -> None:
        names, starts, ends = RangeConfig.flatten(ranges)

        storage.write(path=RangeStorage.names, data=names)
        storage.write(path=RangeStorage.starts, data=starts)
        storage.write(path=RangeStorage.ends, data=ends)

    @staticmethod
    def read_from_config(parser: ConfigParser) -> List[RangeConfig]:
        sheet = ExcelSheet.ranges

        names: List[str] = parser.get(sheet, RangeExcel.name_)
        names = validate_excel_names(names)

        starts: List[Timestamp] = parser.get(sheet, RangeExcel.start)
        ends: List[Timestamp] = parser.get(sheet, RangeExcel.end)

        starts_timestamps = [convert_date_to_timestamp(start) for start in starts]
        ends_timestamps = [convert_date_to_timestamp(end) for end in ends]

        ranges = RangeConfig.reconstruct(
            names=names,
            starts=starts_timestamps,
            ends=ends_timestamps,
        )

        return ranges



def is_within_interval(
    interval_start: int,
    interval_end: int,
    data_start: int,
    data_end: int,
) -> bool:
    """Check if data timestamps fall within the specified interval.

    Determines whether the given data's start and end timestamps are contained
    within the specified interval. The function will return `True` even if the
    data partially overlaps with the interval.

    It is essential to use the same time unit for all timestamps for consistent results.

    Note:
        - The interval is defined by the `interval_start` (inclusive) and
          `interval_end` (exclusive).
        - The function will return `True` if the data spans across the interval
          boundaries.

    Args:
        interval_start (int): The start timestamp of the interval.
        interval_end (int): The end timestamp of the interval.
        data_start (int): The start timestamp of the data.
        data_end (int): The end timestamp of the data.

    Returns:
        `True` if any part of the data falls within the interval, `False` otherwise.
    """
    if (
        (interval_start <= data_start < interval_end)
        or (interval_start < data_end < interval_end)
        or (data_start <= interval_start and data_end >= interval_end)
    ):
        return True

    return False



from rich import print
from rich.progress import track

from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.common.ComputationUmapStorage import ComputationUmapStorage
from processing.common.MeanDistancesMatrix import MeanDistancesMatrix
from processing.config.bands.BandStorage import BandStorage
from processing.config.extractors.ExtractorStorage import ExtractorStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.config.settings.SettingsStorage import SettingsStorage
from processing.interfaces import MenuCallback
from processing.reducers.UmapReducer import UmapReducer
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.filter_nn_extractors import filter_nn_extractors
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.validate_aggregated import validate_aggregated
from processing.utils.validate_autoclusters import validate_autoclusters
from processing.utils.validate_configuration import validate_configuration
from processing.utils.walk_bands_integrations import walk_bands_integrations


@validate_configuration
@validate_autoclusters  # skipping computing if no autoclusters are requested
@validate_aggregated
def compute_requirements(
    storage: Storage,
    callback: MenuCallback,
):
    # TODO: Add check for aggregated data

    print_action("Requirements computation started!", "start")

    ComputationUmapStorage.delete(storage)
    storage.delete(StoragePath.mean_distances_matrix)

    settings = SettingsStorage.read_from_storage(storage)
    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    extractors = ExtractorStorage.read_from_storage(storage)
    nn_extractors = filter_nn_extractors(extractors)
    aggregated_reduceables = AggregatedReduceable.reconstruct(
        bands=bands,
        integrations=integrations,
        nn_extractors=nn_extractors,
    )

    print(
        f"Computing UMAPs..."
        f" (iterations: {settings.computation_umap_iterations},"
        f" dimensions: {settings.computation_umap_dimensions})"
    )

    for ar in aggregated_reduceables:
        features = ar.read_features_from_storage(storage)
        for computation_index in track(
            range(settings.computation_umap_iterations),
            description=f"Band {ar.band.name}, integration {ar.integration.seconds}",
        ):
            umap = UmapReducer(min_dist=0)
            umap.load(
                dimensions=settings.computation_umap_dimensions,
                seed=None,
                features=features,
            )

            computation_features = umap.calculate()

            ComputationUmapStorage.write(
                storage=storage,
                ar=ar,
                data=computation_features,
                index=computation_index,
            )

    print()
    print("Computing mean distances matrix...")

    for band, integration in walk_bands_integrations(bands, integrations):
        computation_umaps = ComputationUmapStorage.read_from_storage(
            storage=storage,
            band=band,
            integration=integration,
        )

        mdm = MeanDistancesMatrix.calculate(features=computation_umaps)
        path = MeanDistancesMatrix.get_path(band, integration)

        storage.write(
            path=path,
            data=mdm,
            compression=True,
        )

    print_action("Requirements computation completed!", "end")
    invoke_menu(storage, callback)



from typing import List

from processing.lib.VAE_v1 import VAE
from processing.reducers.AbstractReducer import AbstractReducer


class VaeReducer(AbstractReducer):
    def calculate(
        self,
    ) -> List[List[float]]:
        features = self._validate_load()
        vae = VAE(n_components=self._dimensions)
        vae.fit(features)
        reduced_features = vae.transform(features)
        self.set_values(reduced_features)
        return self.values



from typing import List, Union

from pandas import Timestamp

from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.config.files.FileConfig import FileConfig
from processing.config.files.FileSheet import FileSheet
from processing.config.labels.LabelConfig import LabelConfig
from processing.config.settings.SettingsConfig import SettingsConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.convert_date_to_timestamp import convert_date_to_timestamp
from processing.utils.read_files_durations import read_files_durations


class FileStorage:
    names = StoragePath.files_names.value
    timestamps = StoragePath.files_timestamps.value
    sites = StoragePath.files_sites.value
    labels = StoragePath.files_labels.value
    durations = StoragePath.files_durations.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(FileStorage.names)
        storage.delete(FileStorage.timestamps)
        storage.delete(FileStorage.sites)
        storage.delete(FileStorage.labels)
        storage.delete(FileStorage.durations)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            storage.exists_dataset(FileStorage.names)
            and storage.exists_dataset(FileStorage.timestamps)
            and storage.exists_dataset(FileStorage.sites)
            and storage.exists_dataset(FileStorage.labels)
            and storage.exists_dataset(FileStorage.durations)
        )

    @staticmethod
    def read_from_storage(
        storage: Storage,
        settings: SettingsConfig,
    ) -> List[FileConfig]:
        names_dataset = storage.read(FileStorage.names)
        names = storage.convert_dataset_to_string_list(names_dataset)

        timestamps = storage.read(FileStorage.timestamps)

        sites_dataset = storage.read(FileStorage.sites)
        sites = storage.convert_dataset_to_string_list(sites_dataset)

        labels_dataset = storage.read(FileStorage.labels)

        # INFO: As labels can be empty, this check is needed
        if labels_dataset.shape[1] == 0:
            labels = []
        else:
            labels = list(list(sublist) for sublist in labels_dataset.asstr()[:])

        durations = storage.read(FileStorage.durations)

        files = FileConfig.reconstruct(
            names=names,
            timestamps=timestamps[:],
            sites=sites,
            labels=labels,
            durations=durations[:],
            audio_path=settings.audio_path,
        )

        return files

    @staticmethod
    def write_to_storage(files: List[FileConfig], storage: Storage) -> None:
        (
            names,
            timestamps,
            sites,
            labels,
            durations,
        ) = FileConfig.flatten(files)

        storage.write(
            path=StoragePath.files_names,
            data=names,
        )

        storage.write(
            path=StoragePath.files_timestamps,
            data=timestamps,
            compression=True,
            attributes={"unit": "milliseconds"},
        )

        storage.write(
            path=StoragePath.files_sites,
            data=sites,
        )

        storage.write(
            path=StoragePath.files_labels,
            data=labels,
        )

        storage.write(
            path=StoragePath.files_durations,
            data=durations,
            attributes={"unit": "milliseconds"},
        )

    @staticmethod
    def read_from_config(
        parser: ConfigParser,
        labels: List[LabelConfig],
        settings: SettingsConfig,
    ) -> List[FileConfig]:
        sheet = ExcelSheet.files

        names = parser.get(sheet, FileSheet.name_)

        dates: List[Union[Timestamp, str]] = parser.get(sheet, FileSheet.date)
        timestamps = [convert_date_to_timestamp(d) for d in dates]

        sites = parser.get(sheet, FileSheet.site)

        labels_values = LabelConfig.convert_to_values_by_file(labels)

        durations = read_files_durations(names, settings.audio_path)

        files = FileConfig.reconstruct(
            names=names,
            timestamps=timestamps,
            sites=sites,
            labels=labels_values,
            durations=durations,
            audio_path=settings.audio_path,
        )

        return files



import os

from processing.config.ConfigParser import ConfigParser
from processing.config.settings.SettingsStorage import SettingsStorage
from processing.utils.get_absolute_path import get_absolute_path


def read_audio_path_from_config(config_path: str) -> str:
    folder_path = os.path.dirname(config_path)
    parser = ConfigParser(config_path, folder_path)
    settings = SettingsStorage.read_from_config(parser)
    return get_absolute_path(settings.audio_path)



from typing import List

from sklearn.preprocessing import robust_scale

from processing.reducers.AbstractReducer import AbstractReducer


class UmapReducer(AbstractReducer):
    min_dist: float

    def __init__(
        self,
        min_dist: float = 0.1,
    ):
        self.min_dist = min_dist

    def calculate(
        self,
    ) -> List[List[float]]:
        features = self._validate_load()
        scaled_features = robust_scale(features)

        # using dynamic import to prevent numba's AOT
        from umap.umap_ import UMAP

        umap = UMAP(
            n_components=self._dimensions,
            random_state=self._seed,
            metric="manhattan",
            min_dist=self.min_dist,
        )

        reduced_features = umap.fit_transform(scaled_features)
        self.set_values(list(reduced_features))  # type: ignore
        return self.values



from typing import Dict, List, Tuple, Type

from processing.config.settings.SettingsConfig import SettingsConfig
from processing.extractors.AcousticComplexityExtractor import (
    AcousticComplexityExtractor,
)
from processing.extractors.AcousticDiversityIndexIndicator import (
    AcousticDiversityIndexExtractor,
)
from processing.extractors.BioacousticsIndexExtractor import BioacousticsIndexExtractor
from processing.extractors.Extractor import Extractor
from processing.extractors.FrequencyEntropyIndicator import FrequencyEntropyExtractor
from processing.extractors.LeqMaadExtractor import LeqMaadExtractor
from processing.extractors.SoundscapeIndexExtractor import SoundscapeIndexExtractor
from processing.extractors.TemporalEntropyExtractor import TemporalEntropyExtractor
from processing.extractors.TemporalMedianExtractor import TemporalMedianExtractor
from processing.extractors.VggExtractor import VggExtractor


# TODO: Add `yamnet` and `YamnetExtractor`
class ExtractorConfig:
    extractors: Dict[str, Type[Extractor]] = {
        "vgg": VggExtractor,
        "leq_maad": LeqMaadExtractor,
        "ht": TemporalEntropyExtractor,
        "med": TemporalMedianExtractor,
        "ndsi": SoundscapeIndexExtractor,
        "aci": AcousticComplexityExtractor,
        "adi": AcousticDiversityIndexExtractor,
        "bi": BioacousticsIndexExtractor,
        "hf": FrequencyEntropyExtractor,
        # TODO: Adapt leq enes output before adding here
    }

    nn_extractors = ["vgg"]
    """The list of extractors using neural networks.

    Add extractor key to this list in order to enable reduction and export features.
    """

    def __init__(
        self,
        index: int,
        name: str,
        offset: int,
        step: int,
        persist: bool,
    ) -> None:
        self.validate_name(name)

        self.index = index
        self.name = name
        self.offset = offset
        self.step = step
        self.persist = persist

    def validate_name(self, name: str) -> None:
        assert name in self.extractors.keys(), f"Unable to find extractor name {name}"

    def instanciate(self, settings: SettingsConfig) -> Extractor:
        instance = self.extractors[self.name]()

        instance.index = self.index
        instance.expected_sample_rate = settings.expected_sample_rate
        instance.offset = self.offset
        instance.step = self.step

        if self.persist:
            instance.persist()

        return instance

    @staticmethod
    def flatten(
        extractors: List["ExtractorConfig"],
    ) -> Tuple[List[str], List[int], List[int], List[int]]:
        names = [e.name for e in extractors]
        offsets = [e.offset for e in extractors]
        steps = [e.step for e in extractors]
        persists = [1 if e.persist is True else 0 for e in extractors]

        return names, offsets, steps, persists

    @staticmethod
    def reconstruct(
        names: List[str],
        offsets: List[int],
        steps: List[int],
        persists: List[bool],
    ) -> List["ExtractorConfig"]:
        extractors: List["ExtractorConfig"] = []

        for index, name in enumerate(names):
            extractor = ExtractorConfig(
                index=index,
                name=name,
                offset=offsets[index],
                step=steps[index],
                persist=persists[index],
            )

            extractors.append(extractor)

        return extractors



from processing.config.files.FileConfig import FileConfig
from processing.common.Interval import Block, Interval


class FileWalker:
    def __init__(
        self,
        file: FileConfig,
        origin: int,
        step: int,
        debug: bool = False,
    ) -> None:
        self.file = file
        self.origin = origin
        self.step = step
        self.debug = debug

        self.load()

    def load(self):
        span = (self.file.end - self.file.start) // self.step

        if self.file.duration < self.step:
            span += 1

        before_remainder = (self.file.start - self.origin) % self.step

        self.before = self.file.start - before_remainder
        self.after = self.before + span * self.step

        if self.file.end > self.after:
            span += 1

        self.span = span

        if self.debug:
            print(
                self.file.index,
                self.span,
                "-",
                self.file.start,
                self.file.end,
                "-",
                self.before,
                self.after,
            )

    def walk(self):
        for i in range(self.span):
            interval_start = self.before + i * self.step
            interval_end = interval_start + self.step

            interval = Interval(
                start=interval_start,
                end=interval_end,
            )

            block = Block(
                file=self.file,
                start=self.file.start if i == 0 else interval_start,
                end=self.file.end if self.file.end < interval_end else interval_end,
            )

            yield interval, block



import numpy as np

from processing.common.MeanDistancesMatrix import MeanDistancesMatrix
from processing.config.Config import Config
from processing.config.bands.BandStorage import BandStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.utils.ask_band import ask_band
from processing.utils.ask_integration import ask_integration
from processing.utils.ask_npy_path import ask_npy_path
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.validate_configuration_with_config import (
    validate_configuration_with_config,
)


@validate_configuration_with_config
def export_mdm(
    config: Config,
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Mean distances matrix export started!", "start")

    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)

    band = ask_band(bands)
    integration = ask_integration(integrations)

    mdm = MeanDistancesMatrix.read_from_storage(storage, band, integration)
    npy_path = ask_npy_path(config)
    np.save(npy_path, mdm)

    print_action("Mean distances matrix export completed!", "end")
    invoke_menu(storage, callback)



from typing import List

from processing.config.autoclusters.AutoclusterConfig import AutoclusterConfig
from processing.config.autoclusters.AutoclusterExcel import AutoclusterExcel
from processing.config.ConfigParser import ConfigParser
from processing.config.ExcelSheet import ExcelSheet
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class AutoclusterStorage:
    names = StoragePath.autoclusters_names.value
    min_cluster_sizes = StoragePath.autoclusters_min_cluster_sizes.value
    min_samples = StoragePath.autoclusters_min_samples.value
    alphas = StoragePath.autoclusters_alphas.value
    epsilons = StoragePath.autoclusters_epsilons.value

    @staticmethod
    def delete_from_storage(storage: Storage) -> None:
        storage.delete(AutoclusterStorage.names)
        storage.delete(AutoclusterStorage.min_cluster_sizes)
        storage.delete(AutoclusterStorage.min_samples)
        storage.delete(AutoclusterStorage.alphas)
        storage.delete(AutoclusterStorage.epsilons)

    @staticmethod
    def exists_in_storage(storage: Storage) -> bool:
        return (
            storage.exists_dataset(AutoclusterStorage.names)
            and storage.exists_dataset(AutoclusterStorage.min_cluster_sizes)
            and storage.exists_dataset(AutoclusterStorage.min_samples)
            and storage.exists_dataset(AutoclusterStorage.alphas)
            and storage.exists_dataset(AutoclusterStorage.epsilons)
        )

    @staticmethod
    def read_from_storage(storage: Storage) -> List[AutoclusterConfig]:
        names_dataset = storage.read(AutoclusterStorage.names)

        names = storage.convert_dataset_to_string_list(names_dataset)
        min_cluster_sizes = storage.read(AutoclusterStorage.min_cluster_sizes)
        min_samples_dataset = storage.read(AutoclusterStorage.min_samples)
        min_samples = storage.convert_dataset_to_string_list(min_samples_dataset)
        alphas = storage.read(AutoclusterStorage.alphas)
        epsilons = storage.read(AutoclusterStorage.epsilons)

        autoclusters = AutoclusterConfig.reconstruct(
            names=names,
            min_cluster_sizes=min_cluster_sizes[:],
            min_samples=min_samples[:],
            alphas=alphas[:],
            epsilons=epsilons[:],
        )

        return autoclusters

    @staticmethod
    def read_from_config(parser: ConfigParser) -> List[AutoclusterConfig]:
        sheet = ExcelSheet.autoclusters

        names = parser.get(sheet, AutoclusterExcel.name_)
        mns = parser.get(sheet, AutoclusterExcel.min_cluster_size)
        min_samples = parser.get(sheet, AutoclusterExcel.min_samples)
        min_samples = [str(ms) for ms in min_samples]
        alphas = parser.get(sheet, AutoclusterExcel.alpha)
        epsilons = parser.get(sheet, AutoclusterExcel.epsilon)

        autoclusters = AutoclusterConfig.reconstruct(
            names=names,
            min_cluster_sizes=mns,
            min_samples=min_samples,
            alphas=alphas,
            epsilons=epsilons,
        )

        return autoclusters

    @staticmethod
    def write_to_storage(
        autoclusters: List[AutoclusterConfig],
        storage: Storage,
    ) -> None:
        (
            names,
            min_cluster_sizes,
            min_samples,
            alphas,
            epsilons,
        ) = AutoclusterConfig.flatten(autoclusters)

        storage.write(path=AutoclusterStorage.names, data=names)
        storage.write(path=AutoclusterStorage.min_cluster_sizes, data=min_cluster_sizes)
        storage.write(path=AutoclusterStorage.min_samples, data=min_samples)
        storage.write(path=AutoclusterStorage.alphas, data=alphas)
        storage.write(path=AutoclusterStorage.epsilons, data=epsilons)



from typing import List

import pandas
from numpy import ndarray


def compute_starting_point(
    paths: List[pandas.DataFrame],
) -> ndarray:
    sum_ = 0

    for path in paths:
        sum_ += path.to_numpy()[0].reshape(1, -1)

    starting_point: ndarray = sum_ / len(paths)  # type: ignore
    return starting_point



from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.config.reducers.ReducerConfig import ReducerConfig
from processing.config.trajectories.TrajectoryConfig import TrajectoryConfig
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath


class TracedStorage:
    @staticmethod
    def delete(storage: Storage) -> None:
        storage.delete(StoragePath.traced)
        storage.delete(StoragePath.traced_timestamps)
        storage.delete(StoragePath.traced_relative_timestamps)

    @staticmethod
    def exists(storage: Storage) -> bool:
        return (
            storage.exists_dataset(StoragePath.traced)
            and storage.exists_dataset(StoragePath.traced_timestamps)
            and storage.exists_dataset(StoragePath.traced_relative_timestamps)
        )

    @staticmethod
    def get_data_path(
        ar: AggregatedReduceable,
        reducer: ReducerConfig,
        trajectory: TrajectoryConfig,
    ) -> str:
        return (
            f"{StoragePath.traced.value}"
            f"/{ar.band.name}"
            f"/{ar.integration.seconds}"
            f"/{ar.extractor.index}"
            f"/{reducer.index}"
            f"/{trajectory.index}"
        )

    @staticmethod
    def write_data(
        storage: Storage,
        trajectory: TrajectoryConfig,
        reducer: ReducerConfig,
        ar: AggregatedReduceable,
    ) -> None:
        path = TracedStorage.get_data_path(ar, reducer, trajectory)

        storage.write(
            path=path,
            data=trajectory.instance.values,
            compression=True,
            attributes={
                "extractor_index": str(ar.extractor.index),
                "reducer_index": str(reducer.index),
                "trajectory_index": str(trajectory.index),
            },
        )

    @staticmethod
    def get_timestamps_path(
        ar: AggregatedReduceable,
        reducer: ReducerConfig,
        trajectory: TrajectoryConfig,
    ) -> str:
        return (
            f"{StoragePath.traced_timestamps.value}"
            f"/{ar.band.name}"
            f"/{ar.integration.seconds}"
            f"/{ar.extractor.index}"
            f"/{reducer.index}"
            f"/{trajectory.index}"
        )

    @staticmethod
    def get_relative_timestamps_path(
        ar: AggregatedReduceable,
        reducer: ReducerConfig,
        trajectory: TrajectoryConfig,
    ) -> str:
        return (
            f"{StoragePath.traced_relative_timestamps.value}"
            f"/{ar.band.name}"
            f"/{ar.integration.seconds}"
            f"/{ar.extractor.index}"
            f"/{reducer.index}"
            f"/{trajectory.index}"
        )

    @staticmethod
    def write_timestamps(
        storage: Storage,
        trajectory: TrajectoryConfig,
        reducer: ReducerConfig,
        ar: AggregatedReduceable,
    ) -> None:
        path = TracedStorage.get_timestamps_path(ar, reducer, trajectory)

        storage.write(
            path=path,
            data=trajectory.instance.timestamps,
            compression=True,
            attributes={
                "extractor_index": str(ar.extractor.index),
                "reducer_index": str(reducer.index),
                "trajectory_index": str(trajectory.index),
            },
        )

    @staticmethod
    def write_relative_timestamps(
        storage: Storage,
        trajectory: TrajectoryConfig,
        reducer: ReducerConfig,
        ar: AggregatedReduceable,
    ) -> None:
        path = TracedStorage.get_relative_timestamps_path(ar, reducer, trajectory)

        storage.write(
            path=path,
            data=trajectory.instance.relative_timestamps,
            compression=True,
            attributes={
                "extractor_index": str(ar.extractor.index),
                "reducer_index": str(reducer.index),
                "trajectory_index": str(trajectory.index),
            },
        )



import numpy as np
from processing.utils.sort_dataframe import sort_dataframe
from sklearn import metrics

from processing.common.MeanDistancesMatrix import MeanDistancesMatrix
from processing.digesters.Digester import Digester


class SilhouetteDigester(Digester):
    def digest(self, labels):
        label = labels[0]
        dataframe, values = self.get_label_data(label)

        silhouette = np.zeros((len(values), len(values)))

        mdm = MeanDistancesMatrix.read_from_storage(
            storage=self.storage,
            band=self.band,
            integration=self.integration,
            trim_half=True,
        )

        for i, cl1 in enumerate(values):
            for j, cl2 in enumerate(values):
                if i > j:
                    labcl1 = np.where(dataframe.index == cl1)[0]
                    labcl1_len = len(labcl1)
                    labcl2 = np.where(dataframe.index == cl2)[0]
                    labcl2_len = len(labcl2)
                    df2 = np.zeros([labcl1_len + labcl2_len, labcl1_len + labcl2_len])
                    df2[:labcl1_len, :labcl1_len] = mdm[np.ix_(labcl1, labcl1)]
                    df2[:labcl1_len, labcl1_len:] = mdm[np.ix_(labcl1, labcl2)]
                    df2[labcl1_len:, labcl1_len:] = mdm[np.ix_(labcl2, labcl2)]
                    df2[labcl1_len:, :labcl1_len] = mdm[np.ix_(labcl2, labcl1)]
                    silhouette[i, j] = metrics.silhouette_score(
                        df2,
                        np.hstack([dataframe.index[labcl1], dataframe.index[labcl2]]),
                        metric="precomputed",
                    )

        silhouette = silhouette + silhouette.T
        silhouette = sort_dataframe(dataframe=silhouette, label=label)
        return silhouette



import numpy as np
from numpy import floating
from numpy.typing import NDArray


def fill_symetrical_matrix_with_nans(matrix: NDArray[floating]) -> NDArray[floating]:
    np.fill_diagonal(matrix, np.nan)
    matrix[np.triu_indices(matrix.shape[0])] = np.nan
    return matrix



from rich.progress import track

from processing.common.AggregatedLabelStorage import AggregatedLabelStorage
from processing.common.AggregatedReduceable import AggregatedReduceable
from processing.config.bands.BandStorage import BandStorage
from processing.config.digesters.DigesterStorage import DigesterStorage
from processing.config.extractors.ExtractorStorage import ExtractorStorage
from processing.config.integrations.IntegrationStorage import IntegrationStorage
from processing.interfaces import MenuCallback
from processing.storage.Storage import Storage
from processing.storage.StoragePath import StoragePath
from processing.utils.filter_nn_extractors import filter_nn_extractors
from processing.utils.invoke_menu import invoke_menu
from processing.utils.print_action import print_action
from processing.utils.print_digesters import print_digesters
from processing.utils.validate_autoclustered import validate_autoclustered
from processing.utils.validate_configuration import validate_configuration
from processing.utils.validate_digesters import validate_digesters


@validate_configuration
@validate_digesters
@validate_autoclustered
def digest(
    storage: Storage,
    callback: MenuCallback,
):
    print_action("Digestions started!", "start")

    storage.delete(StoragePath.digested)

    digesters = DigesterStorage.read_from_storage(storage)
    print_digesters(digesters)

    bands = BandStorage.read_from_storage(storage)
    integrations = IntegrationStorage.read_from_storage(storage)
    extractors = ExtractorStorage.read_from_storage(storage)
    nn_extractors = filter_nn_extractors(extractors)
    aggregated_reduceables = AggregatedReduceable.reconstruct(
        bands=bands,
        integrations=integrations,
        nn_extractors=nn_extractors,
    )

    for ar in aggregated_reduceables:
        aggregated_features = ar.read_features_from_storage(storage)
        aggregated_labels = AggregatedLabelStorage.read_from_storage(
            storage=storage,
            band=ar.band,
            integration=ar.integration,
            extractor=ar.extractor,
        )

        for digester in track(digesters):
            d = digester.instantiate()
            d.features = aggregated_features
            d.labels = aggregated_labels
            d.storage = storage
            d.band = ar.band
            d.integration = ar.integration

            for data, label_a, label_b in d.walk_labels():
                path = (
                    f"{StoragePath.digested.value}"
                    f"/{ar.band.name}"
                    f"/{ar.integration.seconds}"
                    f"/{digester.index}"
                    f"/{label_a.index}"
                )

                attributes = {
                    "digester_name": digester.name,
                    "label_a_property": label_a.property,
                    "label_a_index": str(label_a.index),
                }

                if label_b is not None:
                    path = f"{path}/{label_b.index}"

                    attributes = {
                        **attributes,
                        **{
                            "label_b_property": label_b.property,
                            "label_b_index": str(label_b.index),
                        },
                    }

                storage.write(
                    path=path,
                    data=data,
                    compression=True,
                    attributes=attributes,
                )

    print_action("Digestions completed!", "end")
    invoke_menu(storage, callback)
